import refinitiv.data as rd
import inspect

from refinitiv.data.delivery._data._request import Request
from tests.unit.conftest import StubResponse, StubSession

_defns_with_ext_params = [
    rd.content.custom_instruments.Definition,
    rd.content.custom_instruments.events.Definition,
    rd.content.custom_instruments.search.Definition,
    rd.content.custom_instruments.summaries.Definition,
    rd.content.estimates.view_actuals.annual.Definition,
    rd.content.estimates.view_actuals.interim.Definition,
    rd.content.estimates.view_actuals_kpi.annual.Definition,
    rd.content.estimates.view_actuals_kpi.interim.Definition,
    rd.content.estimates.view_summary.annual.Definition,
    rd.content.estimates.view_summary.historical_snapshots_non_periodic_measures.Definition,
    rd.content.estimates.view_summary.historical_snapshots_periodic_measures_annual.Definition,
    rd.content.estimates.view_summary.historical_snapshots_periodic_measures_interim.Definition,
    rd.content.estimates.view_summary.historical_snapshots_recommendations.Definition,
    rd.content.estimates.view_summary.interim.Definition,
    rd.content.estimates.view_summary.non_periodic_measures.Definition,
    rd.content.estimates.view_summary.recommendations.Definition,
    rd.content.estimates.view_summary_kpi.annual.Definition,
    rd.content.estimates.view_summary_kpi.historical_snapshots_kpi.Definition,
    rd.content.estimates.view_summary_kpi.interim.Definition,
    rd.content.fundamental_and_reference.Definition,
    rd.content.historical_pricing.events.Definition,
    rd.content.historical_pricing.summaries.Definition,
    rd.content.ipa.curves.forward_curves.Definition,
    rd.content.ipa.curves.zc_curve_definitions.Definition,
    rd.content.ipa.curves.zc_curves.Definition,
    rd.content.ipa.dates_and_calendars.add_periods.Definition,
    rd.content.ipa.dates_and_calendars.count_periods.Definition,
    rd.content.ipa.dates_and_calendars.date_schedule.Definition,
    rd.content.ipa.dates_and_calendars.holidays.Definition,
    rd.content.ipa.dates_and_calendars.is_working_day.Definition,
    rd.content.ipa.financial_contracts.bond.Definition,
    rd.content.ipa.financial_contracts.cap_floor.Definition,
    rd.content.ipa.financial_contracts.cds.Definition,
    rd.content.ipa.financial_contracts.cross.Definition,
    rd.content.ipa.financial_contracts.option.Definition,
    rd.content.ipa.financial_contracts.repo.Definition,
    rd.content.ipa.financial_contracts.swap.Definition,
    rd.content.ipa.financial_contracts.swaption.Definition,
    rd.content.ipa.financial_contracts.term_deposit.Definition,
    rd.content.ipa.surfaces.cap.Definition,
    rd.content.ipa.surfaces.eti.Definition,
    rd.content.ipa.surfaces.fx.Definition,
    rd.content.ipa.surfaces.swaption.Definition,
    rd.content.news.headlines.Definition,
    rd.content.ownership.consolidated.breakdown.Definition,
    rd.content.ownership.consolidated.concentration.Definition,
    rd.content.ownership.consolidated.investors.Definition,
    rd.content.ownership.consolidated.recent_activity.Definition,
    rd.content.ownership.consolidated.shareholders_history_report.Definition,
    rd.content.ownership.consolidated.shareholders_report.Definition,
    rd.content.ownership.consolidated.top_n_concentration.Definition,
    rd.content.ownership.fund.breakdown.Definition,
    rd.content.ownership.fund.concentration.Definition,
    rd.content.ownership.fund.holdings.Definition,
    rd.content.ownership.fund.investors.Definition,
    rd.content.ownership.fund.recent_activity.Definition,
    rd.content.ownership.fund.shareholders_history_report.Definition,
    rd.content.ownership.fund.shareholders_report.Definition,
    rd.content.ownership.fund.top_n_concentration.Definition,
    rd.content.ownership.insider.shareholders_report.Definition,
    rd.content.ownership.insider.transaction_report.Definition,
    rd.content.ownership.investor.holdings.Definition,
    rd.content.ownership.org_info.Definition,
    rd.content.pricing.chain.Definition,
    rd.content.pricing.Definition,
    rd.content.search.Definition,
    rd.content.search.lookup.Definition,
    rd.content.symbol_conversion.Definition,
    rd.content.trade_data_service.Definition,
    rd.delivery.omm_stream.Definition,
    rd.delivery.rdp_stream.Definition,
]

_defns_without_ext_params = [
    rd.content.esg.basic_overview.Definition,
    rd.content.esg.bulk.Definition,
    rd.content.esg.full_measures.Definition,
    rd.content.esg.full_scores.Definition,
    rd.content.esg.standard_measures.Definition,
    rd.content.esg.standard_scores.Definition,
    rd.content.esg.universe.Definition,
    rd.content.filings.retrieval.Definition,
    rd.content.filings.search.Definition,
    rd.content.ipa.curves.forward_curves.Definitions,
    rd.content.ipa.curves.zc_curve_definitions.Definitions,
    rd.content.ipa.curves.zc_curves.Definitions,
    rd.content.ipa.dates_and_calendars.add_periods.Definitions,
    rd.content.ipa.dates_and_calendars.count_periods.Definitions,
    rd.content.ipa.dates_and_calendars.holidays.Definitions,
    rd.content.ipa.dates_and_calendars.is_working_day.Definitions,
    rd.content.ipa.financial_contracts.Definitions,
    rd.content.ipa.surfaces.Definitions,
    rd.content.news.story.Definition,
    rd.content.search.metadata.Definition,
    rd.delivery.cfs.buckets.Definition,
    rd.delivery.cfs.file_downloader.Definition,
    rd.delivery.cfs.file_sets.Definition,
    rd.delivery.cfs.files.Definition,
    rd.delivery.cfs.packages.Definition,
    rd.delivery.endpoint_request.Definition,
    rd.session.Definition,
    rd.session.desktop.Definition,
    rd.session.platform.Definition,
]

_defns = _defns_without_ext_params + _defns_with_ext_params

_count_no_has_ext_params = 0
_count_has_ext_params = 0


def write_count():
    global _count_has_ext_params, _count_no_has_ext_params

    for definition in _defns:
        arg_spec = inspect.getfullargspec(definition.__init__)

        if "extended_params" in arg_spec.args:
            _count_has_ext_params += 1
        else:
            _count_no_has_ext_params += 1


def test_definitions_on_extended_params():
    write_count()

    assert _count_has_ext_params == 71, _count_has_ext_params
    assert _count_no_has_ext_params == 29, _count_no_has_ext_params


def test_custom_instruments_definition():
    # region response_400
    response_400 = StubResponse(
        status_code=400,
        content_data=[
            {
                "state": {
                    "code": 400,
                    "status": "Bad Request",
                    "message": "Validation Error",
                },
                "data": [
                    {
                        "key": "symbol",
                        "reason": ".UUID suffix UUID not matched with userID GESG1-60983",
                    }
                ],
            }
        ],
    )
    # endregion
    # region response_200
    response_200 = StubResponse(
        [
            {
                "data": [
                    ["2022-08-26T07:16:54.286Z", 1.9908],
                    ["2022-08-26T07:16:49.777Z", 1.991],
                    ["2022-08-26T07:16:49.297Z", 1.991],
                    ["2022-08-26T07:16:48.796Z", 1.9908],
                    ["2022-08-26T07:16:48.257Z", 1.991],
                    ["2022-08-26T07:16:47.782Z", 1.9908],
                    ["2022-08-26T07:16:47.495Z", 1.991],
                ],
                "defaultPricingField": "TRDPRC_1",
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"decimalChar": ".", "name": "TRDPRC_1", "type": "number"},
                ],
                "interval": None,
                "summaryTimestampLabel": None,
                "universe": {"ric": "S)MyInstrumentEUR.GESG1-60983"},
            }
        ]
    )
    # endregion
    expected_url_usd = "test_get_rdp_url_root/data/custom-instruments/v1/events/S%29MyInstrumentUSD.GESG1-60983?start=2022-05-06T05%3A00%3A00.000000000Z&count=7"
    expected_url_eur = "test_get_rdp_url_root/data/custom-instruments/v1/events/S%29MyInstrumentEUR.GESG1-60983?start=2022-05-06T05%3A00%3A00.000000000Z&count=7"

    def mock_http_request(request: Request):
        testing_url = request.url
        response = response_400
        if "MyInstrumentUSD" in testing_url:
            assert testing_url == expected_url_usd
            response = response_200
        elif "MyInstrumentEUR" in testing_url:
            assert testing_url == expected_url_eur
            response = response_200
        return response

    session = StubSession(is_open=True)
    session.http_request = mock_http_request
    rd.content.custom_instruments.events.Definition(
        universe=["S)MyInstrumentUSD", "S)MyInstrumentEUR"],
        extended_params={
            "start": "2022-05-06T05:00:00.000000000Z",
            "count": 7,
        },
    ).get_data(session)


def test_fundamental_and_reference_definition_udf():
    # region expected_json
    expected_json = {
        "Entity": {
            "E": "DataGrid_StandardAsync",
            "W": {
                "requests": [
                    {
                        "instruments": ["GOOG.O", "AAPL.O"],
                        "fields": [
                            {"name": "TR.Revenue"},
                            {"name": "TR.Revenue.date"},
                            {"name": "TR.GrossProfit"},
                        ],
                        "parameters": {
                            "Scale": 6,
                            "SDate": 0,
                            "EDate": -3,
                            "FRQ": "FY",
                            "Curn": "EUR",
                        },
                        "universe": ["GOOG.O", "AAPL.O"],
                    }
                ]
            },
        }
    }

    # endregion

    def mock_http_request(request: Request):
        testing_json = request.json
        assert testing_json == expected_json
        return response_200

    # region response_200
    response_200 = StubResponse(
        {
            "responses": [
                {
                    "links": {"count": 8},
                    "variability": "",
                    "universe": [
                        {
                            "Instrument": "GOOG.O",
                            "Company Common Name": "Alphabet Inc",
                            "Organization PermID": "5030853586",
                            "Reporting Currency": "USD",
                        },
                        {
                            "Instrument": "AAPL.O",
                            "Company Common Name": "Apple Inc",
                            "Organization PermID": "4295905573",
                            "Reporting Currency": "USD",
                        },
                    ],
                    "data": [
                        ["GOOG.O", 226632.96342, "2021-12-31T00:00:00", 129044.36268],
                        ["GOOG.O", 149453.1076, "2020-12-31T00:00:00", 80074.546],
                        ["GOOG.O", 144386.15542, "2019-12-31T00:00:00", 80250.60966],
                        ["GOOG.O", 119295.22248, "2018-12-31T00:00:00", 67373.2584],
                        ["AAPL.O", 312290.65656, "2021-09-25T00:00:00", 130473.03648],
                        ["AAPL.O", 236041.72275, "2020-09-26T00:00:00", 90246.4166],
                        ["AAPL.O", 237861.47776, "2019-09-28T00:00:00", 89953.90208],
                        ["AAPL.O", 228802.12465, "2018-09-29T00:00:00", 87731.24333],
                    ],
                    "messages": {
                        "codes": [
                            [-1, -1, -1, -1],
                            [-1, -1, -1, -1],
                            [-1, -1, -1, -1],
                            [-1, -1, -1, -1],
                            [-1, -1, -1, -1],
                            [-1, -1, -1, -1],
                            [-1, -1, -1, -1],
                            [-1, -1, -1, -1],
                        ],
                        "descriptions": [{"code": -1, "description": "ok"}],
                    },
                    "headers": [
                        {
                            "name": "instrument",
                            "title": "Instrument",
                            "type": "string",
                            "description": "The requested Instrument as defined by the user.",
                        },
                        {
                            "name": "TR.Revenue",
                            "title": "Revenue",
                            "type": "number",
                            "decimalChar": ".",
                        },
                        {
                            "name": "TR.Revenue",
                            "title": "Date",
                            "type": "datetime",
                        },
                        {
                            "name": "TR.GrossProfit",
                            "title": "Gross Profit",
                            "type": "number",
                            "decimalChar": ".",
                        },
                    ],
                }
            ]
        }
    )
    # endregion

    session = StubSession(is_open=True)
    session.http_request = mock_http_request
    rd.content.fundamental_and_reference.Definition(
        universe=["MSFT.O", "FB.O", "AMZN.O", "TWTR.K"],
        fields=["TR.Revenue", "TR.Revenue.date", "TR.GrossProfit"],
        parameters={"Scale": 6, "SDate": 0, "EDate": -3, "FRQ": "FY", "Curn": "EUR"},
        use_field_names_in_headers=True,
        extended_params={"universe": ["GOOG.O", "AAPL.O"]},
    ).get_data(session)


def test_historical_pricing_events_definition():
    # region response_200
    response = StubResponse(
        [
            {
                "universe": {"ric": "VOD.L"},
                "adjustments": ["exchangeCorrection", "manualCorrection"],
                "defaultPricingField": "TRDPRC_1",
                "qos": {"timeliness": "delayed"},
                "headers": [
                    {"name": "DATE_TIME", "type": "string"},
                    {"name": "EVENT_TYPE", "type": "string"},
                    {"name": "RTL", "type": "number", "decimalChar": "."},
                    {"name": "SEQNUM", "type": "string"},
                    {"name": "BID", "type": "number", "decimalChar": "."},
                    {"name": "BIDSIZE", "type": "number", "decimalChar": "."},
                    {"name": "ASK", "type": "number", "decimalChar": "."},
                    {"name": "ASKSIZE", "type": "number", "decimalChar": "."},
                    {"name": "MID_PRICE", "type": "number", "decimalChar": "."},
                    {"name": "IMB_SH", "type": "number", "decimalChar": "."},
                    {"name": "IMB_SIDE", "type": "string", "isEnum": True},
                    {"name": "QUALIFIERS", "type": "string"},
                ],
                "data": [
                    [
                        "2022-07-19T15:52:52.900000000Z",
                        "quote",
                        50624,
                        "9076320",
                        130.82,
                        12365,
                        131,
                        2153541,
                        130.91,
                        None,
                        None,
                        "[BID_TONE]",
                    ],
                    [
                        "2022-07-19T15:40:07.641000000Z",
                        "quote",
                        50544,
                        "9075948",
                        130.94,
                        80476,
                        131,
                        2153541,
                        130.97,
                        None,
                        None,
                        "[ASK_TONE]",
                    ],
                    [
                        "2022-07-19T15:40:00.158000000Z",
                        "quote",
                        50528,
                        "9075850",
                        130.94,
                        80476,
                        131,
                        2203541,
                        130.97,
                        None,
                        None,
                        "[ASK_TONE]",
                    ],
                    [
                        "2022-07-19T15:39:38.390000000Z",
                        "quote",
                        50496,
                        "9075694",
                        130.94,
                        80476,
                        131,
                        2503541,
                        130.97,
                        None,
                        None,
                        "[ASK_TONE]",
                    ],
                    [
                        "2022-07-19T15:39:01.320000000Z",
                        "quote",
                        50480,
                        "9075524",
                        130.94,
                        80476,
                        131,
                        2453541,
                        130.97,
                        None,
                        None,
                        "[ASK_TONE]",
                    ],
                ],
            }
        ]
    )
    # endregion
    expected_url_goog = "test_get_rdp_url_root/data/historical-pricing/v1/views/events/GOOG.O?eventTypes=quote&count=5&start=2022-07-18T10%3A00%3A00.000000000Z&end=2022-07-19T16%3A00%3A00.000000000Z"
    expected_url_vod = "test_get_rdp_url_root/data/historical-pricing/v1/views/events/VOD.L?eventTypes=quote&count=5&start=2022-07-18T10%3A00%3A00.000000000Z&end=2022-07-19T16%3A00%3A00.000000000Z"

    def mock_http_request(request: Request):
        testing_url = request.url
        if "GOOG" in testing_url:
            assert testing_url == expected_url_goog
        elif "VOD" in testing_url:
            assert testing_url == expected_url_vod
        return response

    session = StubSession(is_open=True)
    session.http_request = mock_http_request
    rd.content.historical_pricing.events.Definition(
        universe=["VOD.L", "GOOG.O"],
        eventTypes="quote",
        extended_params={
            "count": 5,
            "start": "2022-07-18T10:00:00",
            "end": "2022-07-19T16:00:00",
        },
    ).get_data(session)


def test_historical_pricing_summaries_definition():
    # region response_200
    response = StubResponse(
        [
            {
                "universe": {"ric": "EUR="},
                "interval": "P1D",
                "summaryTimestampLabel": "endPeriod",
                "adjustments": [
                    "exchangeCorrection",
                    "manualCorrection",
                    "CCH",
                    "CRE",
                    "RTS",
                    "RPO",
                ],
                "defaultPricingField": "BID",
                "headers": [
                    {"name": "DATE", "type": "string"},
                    {"name": "BID", "type": "number", "decimalChar": "."},
                    {"name": "ASK", "type": "number", "decimalChar": "."},
                ],
                "data": [
                    ["2022-01-05", 1.1313, 1.1315],
                    ["2022-01-04", 1.1285, 1.1289],
                    ["2022-01-03", 1.1294, 1.1298],
                    ["2021-12-31", 1.1368, 1.1372],
                    ["2021-12-30", 1.1323, 1.1327],
                ],
                "status": {
                    "code": "TS.Interday.UserRequestError.70007",
                    "message": "The universe does not support the following fields: [MKT_OPEN].",
                },
                "meta": {
                    "blendingEntry": {
                        "headers": [
                            {"name": "DATE", "type": "string"},
                            {"name": "BID", "type": "number", "decimalChar": "."},
                            {"name": "ASK", "type": "number", "decimalChar": "."},
                        ],
                        "data": [["2022-01-05", 1.1313, 1.1315]],
                    }
                },
            }
        ]
    )
    # endregion
    expected_url_eur = "test_get_rdp_url_root/data/historical-pricing/v1/views/interday-summaries/EUR%3D?start=2021-12-30T00%3A10%3A00.000000000Z&end=2022-01-05T05%3A10%3A00.000000000Z&interval=P1D&adjustments=CCH&count=5&fields=DATE%2CMKT_OPEN%2CBID%2CASK"
    expected_url_gbp = "test_get_rdp_url_root/data/historical-pricing/v1/views/interday-summaries/GBP%3D?start=2021-12-30T00%3A10%3A00.000000000Z&end=2022-01-05T05%3A10%3A00.000000000Z&interval=P1D&adjustments=CCH&count=5&fields=DATE%2CMKT_OPEN%2CBID%2CASK"

    def mock_http_request(request: Request):
        testing_url = request.url
        if "EUR" in testing_url:
            assert testing_url == expected_url_eur
        elif "GBP" in testing_url:
            assert testing_url == expected_url_gbp
        return response

    session = StubSession(is_open=True)
    session.http_request = mock_http_request
    rd.content.historical_pricing.summaries.Definition(
        universe=["EUR=", "GBP="],
        start="2021-12-30T00:10:00",
        end="2022-01-05T05:10:00",
        extended_params={
            "interval": "P1D",
            "adjustments": "CCH",
            "count": 5,
            "fields": "DATE,MKT_OPEN,BID,ASK",
        },
    ).get_data(session)


def test_dates_and_calendars():
    response_200 = StubResponse(
        [
            {"date": "2023-05-30", "holidays": [], "tag": "first req"},
            {"date": "2022-03-30", "holidays": [], "tag": "first req"},
        ]
    )
    expected_json = [
        {
            "startDate": "2020-05-30",
            "tag": "first req",
            "period": "3Y",
            "calendars": ["GER", "UKR"],
            "currencies": ["EUR"],
            "dateMovingConvention": "ModifiedFollowing",
            "endOfMonthConvention": "Same",
            "holidayOutputs": ["Countries", "Names"],
        },
        {
            "startDate": "2022-03-20",
            "tag": "second req",
            "period": "10D",
            "calendars": ["JAP", "UKR"],
            "currencies": ["USD"],
            "dateMovingConvention": "ModifiedFollowing",
            "endOfMonthConvention": "Same",
            "holidayOutputs": ["Countries", "Names"],
        },
    ]

    def mock_http_request(request: Request):
        testing_json = request.json
        assert testing_json == expected_json
        return response_200

    session = StubSession(is_open=True)
    session.http_request = mock_http_request

    def1 = rd.content.ipa.dates_and_calendars.add_periods.Definition(
        tag="my request",
        start_date="2020-04-24",
        period="4D",
        calendars=["BAR", "KOR", "JAP"],
        currencies=["USD"],
        date_moving_convention="NextBusinessDay",
        end_of_month_convention="Last",
        holiday_outputs=["Date", "Calendars", "Names"],
        extended_params={
            "startDate": "2020-05-30",
            "tag": "first req",
            "period": "3Y",
            "calendars": ["GER", "UKR"],
            "currencies": ["EUR"],
            "dateMovingConvention": "ModifiedFollowing",
            "endOfMonthConvention": "Same",
            "holidayOutputs": ["Countries", "Names"],
        },
    )

    def2 = rd.content.ipa.dates_and_calendars.add_periods.Definition(
        tag="my 2 request",
        start_date="2020-04-24",
        period="4D",
        calendars=["BAR", "KOR", "JAP"],
        currencies=["EUR"],
        date_moving_convention="NextBusinessDay",
        end_of_month_convention="Last",
        holiday_outputs=["Date", "Calendars", "Names"],
        extended_params={
            "startDate": "2022-03-20",
            "tag": "second req",
            "period": "10D",
            "calendars": ["JAP", "UKR"],
            "currencies": ["USD"],
            "dateMovingConvention": "ModifiedFollowing",
            "endOfMonthConvention": "Same",
            "holidayOutputs": ["Countries", "Names"],
        },
    )

    defn = rd.content.ipa.dates_and_calendars.add_periods.Definitions([def1, def2])
    defn.get_data(session)


def test_financial_contracts_bond():
    import refinitiv.data.content.ipa.financial_contracts as rdf

    response_200 = StubResponse()
    expected_json = {
        "universe": [
            {
                "instrumentType": "Bond",
                "instrumentDefinition": {
                    "issueDate": "2022-02-28",
                    "endDate": "2032-02-28",
                    "notionalCcy": "USD",
                    "fixedRatePercent": 7,
                    "interestPaymentFrequency": "Annual",
                    "interestCalculationMethod": "Dcb_Actual_Actual",
                },
                "pricingParameters": {"cleanPrice": 155},
            }
        ]
    }

    def mock_http_request(request: Request):
        testing_json = request.json
        assert testing_json == expected_json
        return response_200

    session = StubSession(is_open=True)
    session.http_request = mock_http_request

    definition = rdf.bond.Definition(
        fixed_rate_percent=7,
        interest_calculation_method=rdf.bond.DayCountBasis.DCB_ACTUAL_ACTUAL,
        interest_payment_frequency=rdf.bond.Frequency.ANNUAL,
        pricing_parameters=rdf.bond.PricingParameters(clean_price=122),
        extended_params={
            "instrumentDefinition": {
                "issueDate": "2022-02-28",
                "endDate": "2032-02-28",
                "notionalCcy": "USD",
            },
            "pricingParameters": {"cleanPrice": 155},
        },
    )
    definition.get_data(session)


def test_curves_forward_curves():
    import refinitiv.data.content.ipa.curves as curves

    response_200 = StubResponse()
    expected_json = {
        "universe": [
            {
                "curveDefinition": {
                    "indexName": "EURIBOR",
                    "currency": "EUR",
                    "discountingTenor": "OIS",
                },
                "curveParameters": {"calendarAdjustment": "Calendar"},
                "forwardCurveDefinitions": [
                    {
                        "indexTenor": "3M",
                        "forwardCurveTenors": ["0D", "1D"],
                        "forwardCurveTag": "ForwardTag",
                        "forwardStartDate": "2021-02-01",
                        "forwardStartTenor": "some_start_tenor",
                    }
                ],
                "curveTag": "new_test_curve",
            }
        ]
    }

    def mock_http_request(request: Request):
        testing_json = request.json
        assert testing_json == expected_json
        return response_200

    session = StubSession(is_open=True)
    session.http_request = mock_http_request

    definition = curves.forward_curves.Definition(
        curve_definition=curves.forward_curves.SwapZcCurveDefinition(
            currency="EUR",
            index_name="EURIBOR",
            discounting_tenor="OIS",
        ),
        curve_tag="test_curve",
        extended_params={
            "curveParameters": {"calendarAdjustment": "Calendar"},
            "forwardCurveDefinitions": [
                {
                    "indexTenor": "3M",
                    "forwardCurveTenors": ["0D", "1D"],
                    "forwardCurveTag": "ForwardTag",
                    "forwardStartDate": "2021-02-01",
                    "forwardStartTenor": "some_start_tenor",
                }
            ],
            "curveTag": "new_test_curve",
        },
    )
    definition.get_data(session)


def test_financial_contracts_swap():
    import refinitiv.data.content.ipa.financial_contracts as rdf

    expected_json = {
        "streamID": "5",
        "method": "Subscribe",
        "universe": {
            "instrumentType": "Swap",
            "instrumentDefinition": {
                "tenor": "5Y",
                "legs": [
                    {
                        "direction": "Paid",
                        "interestType": "Float",
                        "notionalCcy": "EUR",
                        "notionalAmount": 1,
                        "interestPaymentFrequency": "Quarterly",
                    },
                    {
                        "direction": "Received",
                        "interestType": "Float",
                        "notionalCcy": "EUR",
                        "indexTenor": "5Y",
                        "interestPaymentFrequency": "Quarterly",
                    },
                ],
            },
            "pricingParameters": {
                "indexConvexityAdjustmentIntegrationMethod": "RiemannSum",
                "indexConvexityAdjustmentMethod": "BlackScholes",
                "valuationDate": "2020-06-01",
            },
        },
        "view": ["InstrumentDescription", "Structure", "Tenor", "InterestType"],
    }

    session = StubSession(is_open=True)

    swap_def = rdf.swap.Definition(
        pricing_parameters=rdf.swap.PricingParameters(
            index_convexity_adjustment_integration_method=rdf.swap.IndexConvexityAdjustmentIntegrationMethod.RIEMANN_SUM,
            index_convexity_adjustment_method=rdf.swap.IndexConvexityAdjustmentMethod.BLACK_SCHOLES,
            valuation_date="2020-06-01",
        ),
        fields=[
            "InstrumentDescription",
            "Structure",
            "Tenor",
            "InterestType",
        ],
        extended_params={
            "instrumentDefinition": {
                "tenor": "5Y",
                "legs": [
                    {
                        "direction": "Paid",
                        "interestType": "Float",
                        "notionalCcy": "EUR",
                        "notionalAmount": 1,
                        "interestPaymentFrequency": "Quarterly",
                    },
                    {
                        "direction": "Received",
                        "interestType": "Float",
                        "notionalCcy": "EUR",
                        "indexTenor": "5Y",
                        "interestPaymentFrequency": "Quarterly",
                    },
                ],
            },
        },
    )
    stream = swap_def.get_stream(session)
    assert stream._stream._stream.open_message == expected_json



__version__ = "1.0.155"
__all__ = (
    "Response",
    "Definitions",
    "bond",
    "cap_floor",
    "cds",
    "cross",
    "option",
    "repo",
    "swap",
    "swaption",
    "term_deposit",
)

from . import bond
from . import cap_floor
from . import cds
from . import cross
from . import option
from . import repo
from . import swap
from . import swaption
from . import term_deposit
from ._definition import Definitions
from ....delivery._data._data_provider import Response



import threading
import time
from unittest.mock import patch

import pytest

from refinitiv.data._content_type import ContentType
from refinitiv.data.delivery._data._api_type import APIType
from refinitiv.data.delivery._stream._protocol_type import ProtocolType
from refinitiv.data.delivery._stream._stream_cxn_cache import (
    StreamCxnCache,
)
from refinitiv.data.delivery._stream._stream_factory import (
    StreamDetails,
)
from refinitiv.data.delivery._stream.event import StreamCxnEvent
from tests.unit.conftest import StubSession, StubResponse
from tests.unit.content.conftest import TIMEOUT
from tests.unit.delivery.stream.conftest import StubStreamConnection


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_stream_cxn_cache_doesnt_close_cxn_when_it_released():
    # given
    from tests.unit.delivery.stream.raw_infos import data

    response = StubResponse(data)
    session = StubSession(fail_if_error=True, response=response)
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )
    cache.get_cxn(session, details)

    # when
    cache.release(session, details)

    # then
    has_cxn = cache.has_cxn(session, details)
    assert has_cxn is True

    # teardown
    cache.close_cxns(session)


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_stream_cxn_cache_close_cxn_correct_and_thread_is_down():
    # given
    from tests.unit.delivery.stream.raw_infos import data

    response = StubResponse(data)
    session = StubSession(fail_if_error=True, response=response)
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )
    cxn = cache.get_cxn(session, details)

    # when
    cache.close_cxns(session)

    time.sleep(TIMEOUT)

    # then
    names = [thread.name for thread in threading.enumerate()]
    assert cxn.name not in names


def test_stream_cxn_cache_not_has_cxn():
    # given
    session = StubSession()
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )

    # when
    has_cxn = cache.has_cxn(session, details)

    # then
    assert not has_cxn


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_stream_cxn_cache_has_cxn():
    # given
    from tests.unit.delivery.stream.raw_infos import data

    response = StubResponse(data)
    session = StubSession(fail_if_error=True, response=response)
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )

    # when
    cache.get_cxn(session, details)

    # then
    has_cxn = cache.has_cxn(session, details)
    assert has_cxn is True

    # teardown
    cache.close_cxns(session)


def test_stream_cxn_cache_returns_empty_list_if_no_cxns():
    # given
    session = StubSession(fail_if_error=True)
    cache = StreamCxnCache()

    # when
    cxns = cache.get_cxns(session)

    # then
    assert len(cxns) == 0


def test_stream_cxn_cache_cannot_release_cxn_if_it_does_not_exists():
    # given
    session = StubSession(fail_if_error=True)
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )

    # then
    with pytest.raises(ValueError, match="because it’s not in the cache"):
        # when
        cache.release(session, details)


def test_stream_cxn_cache_cannot_delete_cxn_if_it_is_none():
    # given
    session = StubSession(fail_if_error=True)
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )

    # then
    with pytest.raises(ValueError, match="because it is empty"):
        # when
        cache.del_cxn(None, session, details)


def test_stream_cxn_cache_cannot_delete_cxn_if_it_does_not_exists():
    # given
    from tests.unit.delivery.stream.raw_infos import data

    response = StubResponse(data)
    session = StubSession(fail_if_error=True, response=response)
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )

    # then
    with pytest.raises(ValueError, match="because already deleted"):
        # when
        cache.del_cxn(StubStreamConnection(), session, details)


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_stream_cxn_cache_cannot_delete_cxn_if_it_is_using():
    # given
    from tests.unit.delivery.stream.raw_infos import data

    response = StubResponse(data)
    session = StubSession(fail_if_error=True, response=response)
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )
    cxn = cache.get_cxn(session, details)

    # then
    with pytest.raises(AssertionError, match="because it is using"):
        # when
        cache.del_cxn(cxn, session, details)

    # teardown
    cache.close_cxns(session)


def test_stream_cxn_cache_has_cxns_is_false():
    # given
    session = StubSession(fail_if_error=True)
    cache = StreamCxnCache()

    # when
    has_cxn = cache.has_cxns(session)

    # then
    assert has_cxn is False


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_stream_cxn_cache_get_cxns():
    # given
    from tests.unit.delivery.stream.raw_infos import data

    response = StubResponse(data)
    session = StubSession(fail_if_error=True, response=response)
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_TRADING
    )
    cache.get_cxn(session, details)
    details = StreamDetails(
        ContentType.STREAMING_CHAINS, ProtocolType.OMM, APIType.STREAMING_PRICING
    )
    cache.get_cxn(session, details)

    # when
    cxns = cache.get_cxns(session)

    # then
    assert len(cxns) == 2

    # teardown
    cache.close_cxns(session)


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_stream_cxn_cache_close_cxns():
    def on_contracts(*args):
        details = StreamDetails(
            ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
        )
        cache.release(session, details)

    def on_chains(*args):
        details = StreamDetails(
            ContentType.STREAMING_CHAINS, ProtocolType.OMM, APIType.STREAMING_PRICING
        )
        cache.release(session, details)

    # given
    from tests.unit.delivery.stream.raw_infos import data

    response = StubResponse(data)
    session = StubSession(fail_if_error=True, response=response)
    cache = StreamCxnCache()
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )
    cxn = cache.get_cxn(session, details)
    cxn.on(StreamCxnEvent.DISCONNECTING, on_contracts)
    details = StreamDetails(
        ContentType.STREAMING_CHAINS, ProtocolType.OMM, APIType.STREAMING_PRICING
    )
    cxn = cache.get_cxn(session, details)
    cxn.on(StreamCxnEvent.DISCONNECTING, on_chains)

    # when
    cache.close_cxns(session)

    # then
    has_cxn = cache.has_cxns(session)
    assert has_cxn is False


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_stream_cxn_cache_get_cxn_async_can_work_asynchronously():
    # given
    from tests.unit.delivery.stream.raw_infos import data

    response = StubResponse(data)
    session = StubSession(response=response)
    details = StreamDetails(
        ContentType.STREAMING_PRICING, ProtocolType.OMM, APIType.STREAMING_PRICING
    )
    cache = StreamCxnCache()

    # when
    try:
        for _ in range(10):
            cache.get_cxn(session, details)
    except Exception as e:
        assert False, str(e)
    else:
        assert True
    finally:
        cache.close_cxns(session)



# coding: utf8

from enum import Enum, unique


@unique
class SurfaceOutputs(Enum):
    HEADERS = "Headers"
    DATATYPE = "DataType"
    DATA = "Data"
    STATUSES = "Statuses"
    FORWARD_CURVE = "ForwardCurve"
    DIVIDENDS = "Dividends"
    INTEREST_RATE_CURVE = "InterestRateCurve"
    GOODNESS_OF_FIT = "GoodnessOfFit"
    UNDERLYING_SPOT = "UnderlyingSpot"
    DISCOUNT_CURVE = "DiscountCurve"
    MONEYNESS_STRIKE = "MoneynessStrike"



import pytest

from refinitiv.data.delivery.cfs import files


@pytest.mark.parametrize(
    "value",
    [
        "get_data",
        "get_data_async",
    ],
)
def test_attribute(value):
    # given

    # when
    definition = files.Definition("")

    # then
    assert hasattr(definition, value)



data = {
    "services": [
        {
            "port": 14002,
            "location": ["ap-southeast-1a"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "ap-southeast-1-aws-1-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [2001, 3000],
        },
        {
            "port": 14002,
            "location": ["ap-southeast-1a", "ap-southeast-1b"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "ap-southeast-1-aws-3-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["ap-southeast-1b"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "ap-southeast-1-aws-2-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["eu-west-1a"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "eu-west-1-aws-1-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["eu-west-1a", "eu-west-1b"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "eu-west-1-aws-3-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["eu-west-1b"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "eu-west-1-aws-2-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["us-east-1a"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "us-east-1-aws-1-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["us-east-1a", "us-east-1b"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "us-east-1-aws-3-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["us-east-1b"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "us-east-1-aws-2-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["us-east-2a"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "us-east-2-aws-1-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["us-east-2a", "us-east-2b"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "us-east-2-aws-3-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 14002,
            "location": ["us-east-2b"],
            "transport": "tcp",
            "provider": "aws",
            "endpoint": "us-east-2-aws-2-med.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["rwf"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["ap-southeast-1a"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "ap-southeast-1-aws-1-lrg.optimized-pricing/api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["ap-southeast-1a", "ap-southeast-1b"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "ap-southeast-1-aws-3-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["ap-southeast-1b"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "ap-southeast-1-aws-2-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["eu-west-1a"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "eu-west-1-aws-1-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["eu-west-1a", "eu-west-1b"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "eu-west-1-aws-3-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["eu-west-1b"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "eu-west-1-aws-2-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["us-east-1a"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "us-east-1-aws-1-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["us-east-1a", "us-east-1b"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "us-east-1-aws-3-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["us-east-1b"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "us-east-1-aws-2-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["us-east-2a"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "us-east-2-aws-1-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["us-east-2a", "us-east-2b"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "us-east-2-aws-3-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
        {
            "port": 443,
            "location": ["us-east-2b"],
            "transport": "websocket",
            "provider": "aws",
            "endpoint": "us-east-2-aws-2-lrg.optimized-pricing-api.refinitiv.net",
            "dataFormat": ["tr_json2"],
            "tier": [1000, 2000],
        },
    ]
}



class Serializer:
    def __init__(self) -> None:
        super().__init__()
        self._dict = {}

    def _get_param(self, name):
        return self._dict.get(name)

    def _set_param(self, name, value):
        if value is None:
            return

        self._dict[name] = value

    def _get_list_param(self, name):
        return self._dict.get(name)

    def _set_list_param(self, name, value):
        if value is None:
            return

        self._dict[name] = [
            item.get_dict() if hasattr(item, "get_dict") else item for item in value
        ]

    def _get_list_of_enums(self, name):
        return self._dict.get(name)

    def _set_list_of_enums(self, enum_type, name, value):
        if value is None:
            return

        self._dict[name] = [
            item.value if isinstance(item, enum_type) else item for item in value
        ]

    def get_dict(self):
        return self._dict



import json
from datetime import datetime

from refinitiv.data.content import news
from tests.unit.conftest import StubSession, StubResponse
from tests.unit.content.news.top_news.data_for_tests import top_news_headlines_df

with open("tests/unit/content/news/top_news/top-news.json", "r") as f:
    top_news_headline_response = json.loads(f.read())


def test_news_top_news_dataframe():
    session = StubSession(
        is_open=True, response=StubResponse(top_news_headline_response)
    )

    response = news.top_news.Definition("").get_data(session=session)
    assert response.data.df.to_string() == top_news_headlines_df


def test_top_news_datetime():
    session = StubSession(
        is_open=True, response=StubResponse(top_news_headline_response)
    )

    response = news.top_news.Definition("").get_data(session=session)
    assert isinstance(response.data.headlines[0].version_created, datetime)
    assert isinstance(response.data.headlines[0].first_created, datetime)



import pandas as pd

page_key_by_column = {
    "revisionId": "revisionId",
    "revisionDate": "revisionDate",
    "topNewsId": "topNewsId",
}


def news_top_hierarchy_build_df(raw: dict, **kwargs) -> pd.DataFrame:
    raw_data = raw.get("data", [{}])

    # data
    columns = page_key_by_column.keys()
    data = [
        [page.get(key) for key in page_key_by_column.values()]
        for category in raw_data
        for page in category.get("pages", [{}])
    ]

    # index
    index_data = [
        (category["name"], page["name"])
        for category in raw_data
        for page in category["pages"]
    ]
    index = pd.MultiIndex.from_tuples(index_data, names=("Category", "Subcategory"))

    return pd.DataFrame(data, index=index, columns=columns)



from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    CrossCurrencyConstituentsDescription,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    FxForwardInstrumentDescription,
    FxSpotInstrumentDescription,
    CrossCurrencyInstrumentDescription,
)


def test_cross_currency_constituents_description():
    # given
    expected_dict = {
        "crossCurrencySwaps": [{}],
        "fxForwards": [{}],
        "fxSpot": {},
        "interestRateSwaps": [{}],
        "overnightIndexSwaps": [{}],
    }

    # when
    testing_obj = CrossCurrencyConstituentsDescription(
        cross_currency_swaps=[CrossCurrencyInstrumentDescription()],
        fx_forwards=[FxForwardInstrumentDescription()],
        fx_spot=FxSpotInstrumentDescription(),
        interest_rate_swaps=[CrossCurrencyInstrumentDescription()],
        overnight_index_swaps=[CrossCurrencyInstrumentDescription()],
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.cross_currency_swaps == [CrossCurrencyInstrumentDescription()]
    assert testing_obj.fx_forwards == [FxForwardInstrumentDescription()]
    assert testing_obj.fx_spot == FxSpotInstrumentDescription()
    assert testing_obj.interest_rate_swaps == [CrossCurrencyInstrumentDescription()]
    assert testing_obj.overnight_index_swaps == [CrossCurrencyInstrumentDescription()]



import json

import allure
import pytest

from refinitiv.data.content import symbol_conversion
from refinitiv.data.content.symbol_conversion import SymbolTypes
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.symbol_conversion.conftest import (
    symbol_conversion_definition,
    invalid_symbol_conversion_definition,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_extended_params_were_sent_in_request,
)


@allure.suite("Content object - Symbol Conversion")
@allure.feature("Content object - Symbol Conversion")
@allure.severity(allure.severity_level.CRITICAL)
class TestSymbolConversion:
    @allure.title("Retrieve symbol conversion with one symbol to mix symbols types")
    @pytest.mark.parametrize(
        "symbols,to_symbol_type,expected_fields",
        [
            (
                "IBM.N",
                [SymbolTypes.OA_PERM_ID, SymbolTypes.TICKER_SYMBOL],
                ["DocumentTitle", "TickerSymbol", "IssuerOAPermID"],
            )
        ],
    )
    @pytest.mark.caseid("C39748890")
    def test_symbol_conversion_ric_to_mix_of_types(
        self, open_session, symbols, to_symbol_type, expected_fields
    ):
        response = symbol_conversion.Definition(
            symbols=symbols,
            from_symbol_type=SymbolTypes.RIC,
            to_symbol_types=to_symbol_type,
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, expected_fields)

    @allure.title("Retrieve symbol conversion with multiple rics to all symbol types")
    @pytest.mark.parametrize(
        "symbols,expected_fields",
        [
            (
                ["AAPL.O", "GOOG.O", "IBM.N"],
                [
                    "RIC",
                    "DocumentTitle",
                    "SEDOL",
                    "TickerSymbol",
                    "IssueISIN",
                    "IssuerOAPermID",
                    "CUSIP",
                ],
            ),
        ],
    )
    @pytest.mark.caseid("4170070")
    def test_symbol_conversion_with_multiple_rics_to_all_types(
        self, open_session, symbols, expected_fields
    ):
        response = symbol_conversion.Definition(
            symbols=symbols, from_symbol_type=SymbolTypes.RIC
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, expected_fields)

    @allure.title("Retrieve symbol conversion to selected symbol types")
    @pytest.mark.parametrize(
        "symbols,to_symbol_type,expected_fields,expected_assets",
        [
            (
                ["US5949181045", "US02079K1079"],
                SymbolTypes.RIC,
                ["RIC", "DocumentTitle"],
                "'Commodities' 'Equities' 'Bond Pricing' 'Funds' 'FX & Money'",
            )
        ],
    )
    @pytest.mark.caseid("C39748891")
    def test_symbol_conversion_from_isin_to_another_two(
        self,
        open_session_async,
        symbols,
        to_symbol_type,
        expected_fields,
        expected_assets,
    ):
        response = symbol_conversion.Definition(
            symbols=symbols,
            from_symbol_type=SymbolTypes.ISIN,
            to_symbol_types=to_symbol_type,
            asset_class=[
                symbol_conversion.AssetClass.COMMODITIES,
                symbol_conversion.AssetClass.EQUITIES,
                symbol_conversion.AssetClass.BONDS,
                symbol_conversion.AssetClass.FUNDS,
                symbol_conversion.AssetClass.WARRANTS,
                symbol_conversion.AssetClass.FX_AND_MONEY,
            ],
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, expected_fields)
        actual_assets = json.loads(response.request_message.content)
        assert expected_assets in actual_assets.get("Filter")

    @pytest.mark.skip("need to refactor")
    @allure.title(
        "Retrieve symbol conversion with valid and invalid params and get data asynchronously"
    )
    @pytest.mark.caseid("C39748892")
    @pytest.mark.smoke
    async def test_symbol_conversion_with_valid_and_invalid_params_and_get_data_async(
        self, open_session_async
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            symbol_conversion_definition(), invalid_symbol_conversion_definition()
        )

        check_response_status(valid_response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.FOUR_HUNDRED,
            expected_http_reason=HttpReason.BAD_REQUEST,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED,
            expected_error_message="Validation error: json.Terms in body must be of type string",
        )

    @pytest.mark.skip("need to refactor")
    @allure.title("Create a invalid symbol conversion definition and get RDError")
    @pytest.mark.caseid("C39748893")
    def test_symbol_conversion_with_invalid_definition(self, open_session):
        with pytest.raises(RDError) as error:
            invalid_symbol_conversion_definition().get_data()
        assert (
            str(error.value)
            == "Error code 400 | Validation error: json.Terms in body must be of type string"
        )

    @allure.title("Retrieve symbol conversion with extended params")
    @pytest.mark.parametrize(
        "symbols,from_symbol_type,to_symbol_type,extended_params,expected_fields",
        [
            (
                ["60000008"],
                SymbolTypes.LIPPER_ID,
                [SymbolTypes.ISIN],
                {"Select": "RIC,SEDOL"},
                ["RIC", "SEDOL"],
            )
        ],
    )
    @pytest.mark.caseid("C39748894")
    def test_symbol_conversion_with_extended_params(
        self,
        open_session,
        symbols,
        from_symbol_type,
        to_symbol_type,
        extended_params,
        expected_fields,
    ):
        response = symbol_conversion.Definition(
            symbols=symbols,
            from_symbol_type=from_symbol_type,
            to_symbol_types=to_symbol_type,
            extended_params=extended_params,
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, expected_fields)
        check_extended_params_were_sent_in_request(response, extended_params)

    @allure.title("Retrieve symbol conversion with closed session")
    @pytest.mark.caseid("C39748895")
    def test_symbol_conversion_with_closed_session(self, open_session):
        session = open_session
        session.close()
        with pytest.raises(ValueError) as error:
            symbol_conversion_definition().get_data()
        assert str(error.value) == "Session is not opened. Can't send any request"



import pandas as pd


def test_response(response):
    assert response


def test_success_response_is_success(response):
    success = response.is_success
    assert success, [
        response.errors,
        response.request_message.url,
    ]
    assert isinstance(success, bool)


def test_not_success_response_is_success(response):
    success = response.is_success
    assert not success
    assert isinstance(success, bool)


def test_success_response_data(response):
    data = response.data
    assert data, f"No data in response,\nresponse.data={data}"
    assert isinstance(
        data.raw, (dict, list)
    ), f"data.raw is not a dict or list it is a {type(data.raw)}"
    assert isinstance(
        data.df, pd.DataFrame
    ), f"data.df is not a DataFrame it is a {type(data.df)}"
    assert not data.df.empty, "Data frame is empty"


def test_success_response_data_empty(response):
    data = response.data
    assert data
    assert isinstance(data.raw, dict)
    assert isinstance(data.df, pd.DataFrame)
    assert data.df.empty


def test_success_response_data_text(response):
    data = response.data
    assert data
    assert isinstance(data.raw, dict)
    assert isinstance(data.text, str)
    assert data.text


def test_not_success_response_data(response):
    data = response.data
    assert data
    assert data.raw
    assert data.df is None or data.df.empty


def test_success_response_status(status):
    assert status
    assert isinstance(status, dict)
    assert status["http_status_code"] == 200, (
        status["error"] if status.get("error") else status
    )
    assert status["http_reason"] == "OK"
    has_content = status.get("content", False)
    assert not has_content or "error" not in status["content"]


def test_not_success_response_status(response):
    status = response.http_status
    assert status
    assert isinstance(status, dict)
    has_content = status.get("content", False)
    has_error = status.get("error", False)
    assert has_content or has_error
    assert (has_content and "error" in status["content"]) or "error" in status


def test_response_headers(response):
    headers = response.http_headers
    assert headers


def test_response_request_message(response):
    request_message = response.request_message
    assert request_message


def test_response_closure(response, mock_closure=None):
    closure = response.closure
    if mock_closure:
        assert closure == mock_closure
    else:
        assert closure is None


def success_response_tests(response, mock_closure=None):
    test_success_response_is_success(response)
    test_response(response)
    test_response_closure(response, mock_closure)
    test_response_request_message(response)
    test_response_headers(response)
    test_success_response_status(response.http_status)
    test_success_response_data(response)


def success_response_text_tests(response, mock_closure=None):
    test_response(response)
    test_response_closure(response, mock_closure)
    test_response_request_message(response)
    test_response_headers(response)
    test_success_response_status(response.http_status)
    test_success_response_data_text(response)
    test_success_response_is_success(response)


def success_response_data_empty_tests(response):
    test_response(response)
    test_response_closure(response)
    test_response_request_message(response)
    test_response_headers(response)
    test_success_response_status(response.http_status)
    test_success_response_data_empty(response)
    test_success_response_is_success(response)


def not_success_response_tests(response, check_status=True):
    test_response(response)
    test_response_closure(response)
    test_response_request_message(response)
    test_response_headers(response)
    check_status and test_not_success_response_status(response)
    test_not_success_response_data(response)
    test_not_success_response_is_success(response)


def test_alignment_response_success(response):
    assert response, f"Responce is empty"
    test_success_response_is_success(response)
    test_success_response_status(response.http_status)
    assert response.http_headers, f"HTTP headers are empty\n{response.http_headers}"
    assert response.data.raw, "Raw data is empty"
    test_success_response_data(response)



import pytest
import pandas as pd

from refinitiv.data.eikon._tools import convert_content_data_to_df_udf


@pytest.mark.parametrize(
    ("input_data",),
    [
        (
            {
                "headlines": [
                    {
                        "firstCreated": "2021-12-29T07:00:44.686Z",
                        "sourceCode": "NS:PRN",
                        "storyId": "urn:newsml:reuters.com:20211229:nPRrSFC87a:1",
                        "text": None,
                        "versionCreated": "2021-12-29T07:00:44.686Z",
                    },
                    {
                        "firstCreated": "2021-12-29T02:27:55.372Z",
                        "sourceCode": "NS:ECLCTA",
                        "storyId": "urn:newsml:reuters.com:20211229:nNRAiknlz9:1",
                        "text": "",
                        "versionCreated": "2021-12-29T02:27:55.372Z",
                    },
                    {
                        "firstCreated": "2021-12-29T02:27:08.256Z",
                        "sourceCode": "NS:ECLCTA",
                        "storyId": "urn:newsml:reuters.com:20211229:nNRAiknks9:1",
                        "text": "Ibm Maintenance And Support Renewal",
                        "versionCreated": "2021-12-29T02:27:08.256Z",
                    },
                ],
            },
        )
    ],
)
def test_convert_content_data_to_df_udf(input_data):
    result = convert_content_data_to_df_udf(input_data)
    assert not result.empty

    for row in result.itertuples():
        if row.storyId == "urn:newsml:reuters.com:20211229:nPRrSFC87a:1":
            assert row.text is pd.NA
        if row.storyId == "urn:newsml:reuters.com:20211229:nNRAiknlz9:1":
            assert row.text == ""
        if row.storyId == "urn:newsml:reuters.com:20211229:nNRAiknks9:1":
            assert row.text == "Ibm Maintenance And Support Renewal"


@pytest.mark.parametrize(
    ("input_data",),
    [
        (
            {
                "headlines": [],
            },
        )
    ],
)
def test_convert_content_data_to_df_udf_empty_df(input_data):
    # given
    # when
    result = convert_content_data_to_df_udf(input_data)
    # then
    assert result.empty


def test_convert_content_data_to_df_udf_convert_data_types():
    pass



import pandas as pd


INPUT_DF_WITHOUT_DATE = pd.DataFrame(
    {
        "IssuerCountry": {0: "ConstantIndex"},
    }
)
EXPECTED_DF_WITHOUT_DATE = INPUT_DF_WITHOUT_DATE


INPUT_DF_WITH_DATE = pd.DataFrame(
    {
        "IssuerCountry": {0: "ConstantIndex"},
        "IssueDate": {0: "2032-02-28 00:00:00+0000"},
    }
)
EXPECTED_DF_WITH_DATE = pd.DataFrame(
    {
        "IssuerCountry": {0: "ConstantIndex"},
        "IssueDate": {0: pd.to_datetime("2032-02-28 00:00:00")},
    }
)



# coding: utf8


from enum import Enum, unique


@unique
class LocalVolatilityMethod(Enum):
    BEST_SMILE = "BestSmile"
    CONVEX_SMILE = "ConvexSmile"
    PARABOLA_SMOOTH = "ParabolaSmooth"
    PARABOLA_WITHOUT_EXTRAPOL = "ParabolaWithoutExtrapol"
    RATIONAL_SMOOTH = "RationalSmooth"
    RATIONAL_WITHOUT_EXTRAPOL = "RationalWithoutExtrapol"



import os
from typing import Optional

from ._external_libraries import python_configuration as ext_config_mod  # noqa
from ._configure import get_config as _get_config, _read_config_file, _RDPConfig

RDConfig = _RDPConfig


def get_config() -> RDConfig:
    """
    Returns
    -------
    config object
    """
    return _get_config()


def load_config(path: Optional[str]) -> RDConfig:
    """
    Load user's config file and set this file as default.

    Parameters
    ----------
        path: str
            Path to user's config file.

    Raises
    ----------
    Exception
        If can't find file by path that user provided

    Returns
    ----------
    config object
    """
    return _load_config_and_set_default(path)


def _load_config_and_set_default(path: Optional[str]) -> _RDPConfig:
    if not os.path.exists(path):
        raise FileNotFoundError(
            f"Can't find file: {path}. Current working folder {os.getcwd()}"
        )

    loaded_config = _read_config_file(path)
    user_config = ext_config_mod.config_from_dict(loaded_config)

    config = _get_config()
    config._set_config_index(0, user_config)
    return config



import warnings
from contextlib import nullcontext as does_not_raise

import pytest

import refinitiv.data._external_libraries.python_configuration as ext_config
from refinitiv.data._core.session._session_cxn_type import SessionCxnType
from refinitiv.data.delivery._stream._stream_cxn_config_provider import (
    CxnConfigProvider,
    PlatformAndDeployedCxnConfigProvider,
    DeployedCxnConfigProvider,
    PlatformCxnConfigProvider,
    DesktopCxnConfigProvider,
    get_cxn_config,
)
from refinitiv.data.errors import RDError
from tests.unit.conftest import (
    StubConfig,
    StubSession,
    args,
    error_403_access_denied_http_response_generator,
    StubResponse,
)


@pytest.mark.parametrize(
    "transport,url,expected",
    [
        args(
            transport="websocket",
            url="/head/middle/v1",
            expected=pytest.raises(ValueError),
        ),
        args(
            transport="websocket",
            url="://amer-3-t3.beta.streaming-pricing-api.refinitiv.com:443",
            expected=pytest.raises(ValueError),
        ),
        args(
            transport="websocket",
            url="ws://:443",
            expected=pytest.raises(ValueError),
        ),
        args(
            transport="websocket",
            url="ws:///head/middle/v1",
            expected=pytest.raises(ValueError),
        ),
        args(
            transport="websocket",
            url="wss://amer-3-t3.beta.streaming-pricing-api.refinitiv.com:443",
            expected=does_not_raise(),
        ),
        args(
            transport="websocket",
            url="10.3.177.158:15000",
            expected=does_not_raise(),
        ),
        args(
            transport="websocket",
            url="//amer-3-t3.beta.streaming-pricing-api.refinitiv.com:443",
            expected=does_not_raise(),
        ),
        args(
            transport="websocket",
            url="amer-3-t3.beta.streaming-pricing-api.refinitiv.com:443",
            expected=does_not_raise(),
        ),
        args(
            transport="websocket",
            url="//amer-3-t3.beta.streaming-pricing-api.refinitiv.com",
            expected=does_not_raise(),
        ),
        args(
            transport="websocket",
            url="amer-3-t3.beta.streaming-pricing-api.refinitiv.com",
            expected=does_not_raise(),
        ),
        args(
            transport="websocket",
            url="head/middle/v1",
            expected=does_not_raise(),
        ),
    ],
)
def test_url_parsed_success(transport, url, expected):
    with expected:
        CxnConfigProvider.info_from_url(transport=transport, url=url)


def test__request_infos_raises_exception_if_you_dont_have_access():
    discovery_url = "just-an-url"
    api_config_key = "just-a-key.endpoints.value"

    config = StubConfig()
    session = StubSession()
    session._response = error_403_access_denied_http_response_generator()

    config_provider = CxnConfigProvider()
    with pytest.raises(RDError):
        config_provider._request_infos(discovery_url, api_config_key, config, session)


@pytest.mark.parametrize(
    "provider_class, expected_num_infos",
    [
        (DesktopCxnConfigProvider, 3),
        (PlatformCxnConfigProvider, 3),
        (DeployedCxnConfigProvider, 1),
        (PlatformAndDeployedCxnConfigProvider, 1),
    ],
)
def test_possibility_to_set_list_of_streaming_urls(provider_class, expected_num_infos):
    config = ext_config.config_from_dict(
        {
            "sessions": {
                "platform": {
                    "default": {
                        "realtime-distribution-system": {"url": "10.3.177.158:15000"}
                    }
                }
            },
            "apis": {
                "streaming": {
                    "pricing": {
                        "endpoints": {
                            "main": {
                                "direct-url": [
                                    "beta.streaming-pricing.com",
                                    "wss://beta.streaming-pricing.com:443",
                                    "10.3.177.158:15000",
                                ],
                                "protocols": ["OMM"],
                                "locations": [],
                            }
                        }
                    }
                }
            },
        }
    )
    session = StubSession(config=config)
    config_provider = provider_class()
    cfg = config_provider.get_cfg(session, "apis.streaming.pricing.endpoints.main")

    assert cfg.num_infos == expected_num_infos


@pytest.mark.parametrize(
    "session_cxn_type, will_warn",
    [
        (SessionCxnType.DEPLOYED, False),
        (SessionCxnType.REFINITIV_DATA, False),
        (SessionCxnType.REFINITIV_DATA_AND_DEPLOYED, False),
        (SessionCxnType.DESKTOP, True),
    ],
)
def test_direct_url_cannot_used_with_desktop_session(session_cxn_type, will_warn):
    # given
    api_cfg_key = "apis.streaming.pricing.endpoints.main"
    config = ext_config.config_from_dict(
        {
            "sessions": {
                "platform": {
                    "default": {
                        "realtime-distribution-system": {"url": "10.3.177.158:15000"}
                    }
                }
            },
            "apis": {
                "streaming": {
                    "pricing": {
                        "url": "/streaming/pricing/v1",
                        "endpoints": {
                            "main": {
                                "direct-url": "wss://beta.streaming-pricing.com:443",
                                "path": "/",
                                "protocols": ["OMM"],
                                "locations": [],
                            }
                        },
                    }
                }
            },
        }
    )
    session = StubSession(
        config=config,
        response=StubResponse(
            {
                "services": [
                    {
                        "dataFormat": ["tr_json2"],
                        "endpoint": "localhost/api/rdp/streaming/pricing/v1/WebSocket",
                        "location": ["local"],
                        "port": 9001,
                        "provider": "local",
                        "transport": "websocket",
                    }
                ]
            }
        ),
    )
    session._session_cxn_type = session_cxn_type

    if will_warn:
        with warnings.catch_warnings(record=True) as w:
            # when
            cfg = get_cxn_config(api_cfg_key, session)

            # then
            assert len(w) == 1
            assert issubclass(w[0].category, UserWarning)
            assert (
                "direct-url config parameter cannot be used with desktop session, "
                f"the '{api_cfg_key}.direct-url' "
                "parameter will be discarded."
            ) == str(w[0].message)

            assert (
                cfg.url == "ws://localhost:9001/api/rdp/streaming/pricing/v1/WebSocket"
            )

    else:
        with warnings.catch_warnings(record=True) as w:
            # when
            get_cxn_config(api_cfg_key, session)

            # then
            assert len(w) == 0, w[0].message



import datetime
import json
import logging
import os

from ....delivery.cfs._tools import remove_one_ext
from .... import _log as log
from ...._tools._common import cached_property


def parse_log_file(logpath: str) -> dict:
    details_by_log_action = {}
    try:
        with open(logpath) as f:
            for line in f:
                log_line = json.loads(line)
                log_action = log_line.get("action")

                if log_action == "CLEANUP":
                    details_by_log_action["UPDATE"] = []
                    continue

                if not log_action:
                    continue

                log_details = log_line.get("details")
                details = details_by_log_action.setdefault(log_action, [])

                if log_action == "CREATE TABLE":
                    detail_item = log_details.get("table-name")
                else:
                    detail_item = log_details.get("filename")

                details.append(detail_item)

    except Exception:
        pass

    return details_by_log_action


def get_file_logger(id_, logpath_):
    file_logger = log.root_logger().getChild(f"{id_}.file")
    handler = logging.FileHandler(logpath_, "a")
    formatter = logging.Formatter("%(message)s")
    handler.setFormatter(formatter)
    file_logger.addHandler(handler)
    file_logger.setLevel(logging.DEBUG)
    file_logger.propagate = False
    return file_logger


class Actions:
    def __init__(self, id_, logpath):
        self.id = id_
        self.logpath = logpath
        self._details_by_action = None

    @property
    def details_by_action(self):
        if self._details_by_action is None:
            self._details_by_action = parse_log_file(self.logpath)
        return self._details_by_action

    @cached_property
    def _file_logger(self):
        return get_file_logger(self.id, self.logpath)

    def _log(self, action_name, details):
        date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
        s = json.dumps({"date": date, "action": action_name, "details": details})
        self._file_logger.info(s)

    def add(self, action_name, **details):
        self._log(action_name, details)
        return details

    def downloaded(self, **details):
        self._log("DOWNLOAD", details)
        return details

    def extracted(self, **details):
        self._log("EXTRACT", details)
        return details

    def table_created(self, **details):
        self._log("CREATE TABLE", details)
        return details

    def updated(self, **details):
        self._log("UPDATE", details)
        return details

    def cleaned_up_db(self, **details):
        self._log("CLEANUP", details)
        return details

    def _get(self, action_name):
        return set(self.details_by_action.get(action_name, []))

    def get_downloaded(self):
        return self._get("DOWNLOAD")

    def get_extracted(self):
        return self._get("EXTRACT")

    def get_created_tables(self):
        return self._get("CREATE TABLE")

    def get_not_updated_and_extracted(self):
        downloaded = set(remove_one_ext(filename) for filename in self.get_downloaded())
        updated = self._get("UPDATE")
        not_updated = downloaded - updated
        return not_updated, self._get("EXTRACT")

    def cleanup(self):
        try:
            os.remove(self.logpath)
        except PermissionError:
            log.root_logger().warning(
                f"Cannot remove {self.logpath}, file will be cleared."
            )
            with open(self.logpath, "w") as f:
                f.write("")
            log.root_logger().warning(f"{self.logpath} cleared.")
        except FileNotFoundError:
            # do nothing
            pass

    def dispose(self):
        log.dispose_logger(self._file_logger)
        self.cleanup()

    def update(self):
        self._details_by_action = None



from .tools import Sensitive


class GrantPassword:
    def __init__(self, username: str, password: str, token_scope: str = "trapi"):
        """
        Parameters
        ----------
        username: str
            The username to use for authentication
        password: str
            The password to use for authentication
        token_scope: str, list of str (default: trapi)
            The scope to use for authentication

        Examples
        --------
        >>> import refinitiv.data as rdp
        >>> rdp.session.platform.GrantPassword(username="username", password="password", token_scope="token_scope")
        or
        >>> import refinitiv.data as rdp
        >>> rdp.session.platform.GrantPassword("username", "password", "token_scope")

        """
        self._token_scope = token_scope
        self._username = username
        self._password = Sensitive(password)

    def get_username(self):
        return self._username

    def username(self, user_name: str):
        self._username = user_name
        return self

    def get_token_scope(self):
        return self._token_scope

    def with_scope(self, token_scope):
        if token_scope:
            self._token_scope = token_scope
        return self

    def get_password(self):
        return self._password

    def password(self, value: str):
        self._password = Sensitive(value)
        return self

    def is_valid(self):
        return self._password and self._username



from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    BidAskFieldsFormulaDescription,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    FieldDescription,
)


def test_bid_ask_fields_formula_description():
    # given
    expected_dict = {"bid": {}, "ask": {}}

    # when
    testing_obj = BidAskFieldsFormulaDescription(
        bid=FieldDescription(),
        ask=FieldDescription(),
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.bid == FieldDescription()
    assert testing_obj.ask == FieldDescription()



import abc
import json
import pathlib
import threading
import warnings
from typing import List, Optional, TYPE_CHECKING, Union, Dict, Type
from urllib.parse import urlsplit

import httpx

from ._stream_cxn_config_data import (
    StreamServiceInfo,
    DesktopStreamCxnConfig,
    PlatformStreamCxnConfig,
    StreamCxnConfig,
)
from .._data import Request, RequestMethod
from ..._configure import keys
from ..._core.session import SessionCxnType
from ..._core.session.tools import get_delays
from ..._errors import ScopeError, RDError
from ..._tools import parse_url, urljoin

if TYPE_CHECKING:
    from ..._core.session import Session, PlatformSession
    from ..._configure import _RDPConfig


def get_discovery_url(
    root_url: str, base_config_name: str, full_config_name: str, config: "_RDPConfig"
) -> str:
    base_path = config.get_str(f"{base_config_name}.url")

    try:
        endpoint_path = config.get_str(f"{full_config_name}.path")
    except KeyError:
        raise KeyError(
            f"Cannot find discovery endpoint '{full_config_name}' into config."
        )

    if base_path.startswith("http"):
        url = base_path
    else:
        url = urljoin(root_url, base_path)

    return urljoin(url, endpoint_path)


def _filter_by_location(locations: List[str], infos: List[StreamServiceInfo]) -> list:
    if not locations:
        return infos

    filtered = []
    for location in locations:
        for info in infos:
            has_location = any(
                loc.strip().startswith(location) for loc in info.location
            )
            if has_location and info not in filtered:
                filtered.append(info)

    return filtered


def create_infos(data, transport, port_by_prefix, tier):
    infos = []
    for service in data.get("services", []):
        if service.get("transport") != transport:
            continue

        if tier is not None and "tier" in service:
            tier_range: List[int] = service["tier"]
            if tier < tier_range[0] or tier > tier_range[1]:
                continue

        endpoint_path = pathlib.Path(service.get("endpoint"))
        host = str(endpoint_path.parts[0])
        path = "/".join(endpoint_path.parts[1:])

        scheme = ""
        port = service.get("port")
        if transport == "websocket":
            scheme = port_by_prefix.get(port, "ws")

        infos.append(
            StreamServiceInfo(
                scheme=scheme,
                host=host,
                port=port,
                path=path,
                data_formats=service.get("dataFormat", ["unknown"]),
                location=service.get("location"),
                transport=transport,
            )
        )
    return infos


def get_base_cfg_name(cfg_key: str) -> str:
    return cfg_key.split(".endpoints.")[0]


class CxnConfigProvider(abc.ABC):
    config_class = None
    _port_by_prefix = {
        80: "ws",
        443: "wss",
    }

    def __init__(self) -> None:
        self._start_connecting = threading.Event()
        self._start_connecting.set()
        self._timer = threading.Event()
        self._delays = get_delays()

    def wait_start_connecting(self):
        self._start_connecting.clear()

    def start_connecting(self):
        self._start_connecting.set()

    def get_cfg(
        self, session: "Session", api_cfg_key: str
    ) -> Union[PlatformStreamCxnConfig, DesktopStreamCxnConfig]:
        """
        Parameters
        ----------
        session: Session
        api_cfg_key: str
            Example - "apis.streaming.pricing.endpoints.main"

        Returns
        -------
        PlatformStreamCxnConfig or DesktopStreamCxnConfig

        """
        base_config_name = get_base_cfg_name(api_cfg_key)
        cfg: "_RDPConfig" = session.config

        if cfg.get(f"{base_config_name}.use_rwf", False):
            transport = "tcp"
        else:
            transport = "websocket"

        urls: Union[str, List[str], None] = None

        if transport in {"tcp", "websocket"}:
            urls = cfg.get(f"{api_cfg_key}.direct-url")

        session_cxn_type = session._get_session_cxn_type()
        if urls is not None and session_cxn_type == SessionCxnType.DESKTOP:
            warnings.warn(
                "direct-url config parameter cannot be used with desktop session, "
                f"the '{api_cfg_key}.direct-url' parameter will be discarded."
            )
            urls = None

        if urls is not None:
            if not isinstance(urls, list):
                urls = [urls]

            infos = [self.info_from_url(transport, url) for url in urls]

        else:
            infos = self._request_infos(
                discovery_url=get_discovery_url(
                    session._get_rdp_url_root(), base_config_name, api_cfg_key, cfg
                ),
                api_cfg_key=api_cfg_key,
                config=cfg,
                session=session,
                transport=transport,
            )

        protocols = cfg.get_list(f"{api_cfg_key}.protocols")
        return self._create_cfg(session, infos, protocols, transport, api_cfg_key)

    @staticmethod
    def info_from_url(transport: str, url: str, data_formats=None) -> StreamServiceInfo:
        if data_formats is None:
            data_formats = ["unknown"]

        # If there is no scheme or netloc add netloc marker to make it valid URL
        if not ("://" in url or url.startswith("//")) and not url.startswith("/"):
            url = "//" + url

        result = parse_url(url)
        scheme = result.scheme
        host = result.hostname
        port = result.port
        path = result.path

        # If url parsing did not get valid hostname, raise exception
        if not host:
            raise ValueError(f"Invalid URL: {url}")

        if not scheme and transport == "websocket":
            scheme = "wss" if port == 443 else "ws"

        return StreamServiceInfo(
            scheme=scheme or "",
            host=host,
            port=port or 80,
            path=path or "",
            data_formats=data_formats,
            location="",
            transport=transport,
        )

    def _request_infos(
        self,
        discovery_url: str,
        api_cfg_key: str,
        config: "_RDPConfig",
        session: "Session",
        transport: str = "websocket",
    ) -> List[StreamServiceInfo]:
        method = RequestMethod.GET
        try:
            session.verify_scope(urlsplit(discovery_url).path, method)
        except ScopeError as e:
            session.error(
                f"Insufficient Scope. Cannot load the list of associated URLs "
                f"from {discovery_url} for {api_cfg_key} endpoint."
            )
            raise e
        tier: Optional[int] = config.get(f"{get_base_cfg_name(api_cfg_key)}.tier")

        response = None
        once = False
        server_mode = session.server_mode
        while not once or server_mode is True:
            once = True
            self._start_connecting.wait()
            try:
                request = Request(
                    url=discovery_url,
                    method=method,
                    # server won't accept tier: false
                    params={"tier": True} if tier else {},
                    auto_retry=True,
                )
                response = session.http_request(request)
            except httpx.HTTPError:
                if server_mode is True:
                    delay = self._delays.next()
                    session.debug(
                        f"CxnConfigProvider waiting {delay} secs "
                        f"until the next attempt."
                    )
                    self._timer.wait(delay)
                else:
                    break
            else:
                break

        try:
            data = response.json()
        except (AttributeError, json.decoder.JSONDecodeError):
            message = (
                f"Cannot load the list of associated URLs "
                f"from {discovery_url} for {api_cfg_key} endpoint."
            )
            session.error(message)
            raise ConnectionError(message) from None

        err = data.get("error")
        if err:
            raise RDError(response.status_code, err.get("message"))

        infos = create_infos(data, transport, self._port_by_prefix, tier)
        return self._filter_infos(infos, api_cfg_key, config)

    def _filter_infos(
        self,
        infos: List[StreamServiceInfo],
        api_cfg_key: str,
        cfg: "_RDPConfig",
    ) -> List[StreamServiceInfo]:
        return infos

    def _create_cfg(
        self,
        session: "Session",
        infos: List[StreamServiceInfo],
        protocols: List[str],
        transport: str = "websocket",
        api_cfg_key: str = "",
    ) -> Union[PlatformStreamCxnConfig, DesktopStreamCxnConfig]:
        return self.config_class(infos, protocols, transport, api_cfg_key)


class DesktopCxnConfigProvider(CxnConfigProvider):
    config_class = DesktopStreamCxnConfig

    def _create_cfg(
        self,
        session: "Session",
        infos: List[StreamServiceInfo],
        protocols: List[str],
        transport: str = "websocket",
        api_cfg_key: str = "",
    ) -> Union[PlatformStreamCxnConfig, DesktopStreamCxnConfig]:
        return self.config_class(session, infos, protocols, api_cfg_key)


class PlatformCxnConfigProvider(CxnConfigProvider):
    config_class = PlatformStreamCxnConfig

    def _filter_infos(
        self,
        infos: List[StreamServiceInfo],
        api_cfg_key: str,
        cfg: "_RDPConfig",
    ) -> List[StreamServiceInfo]:
        locations = cfg.get_list(f"{api_cfg_key}.locations")
        return _filter_by_location(locations, infos)


class DeployedCxnConfigProvider(CxnConfigProvider):
    def get_cfg(
        self, session: "PlatformSession", api_cfg_key: str
    ) -> PlatformStreamCxnConfig:

        url: str = session._deployed_platform_host
        cfg: "_RDPConfig" = session.config

        if url is None:
            session_name: str = session.name
            key = keys.platform_realtime_distribution_system(session_name)
            url_key = f"{key}.url"
            url = cfg.get_str(url_key)

        if cfg.get(f"{get_base_cfg_name(api_cfg_key)}.use_rwf", False):
            transport = "tcp"
            data_formats = None
        else:
            transport = "websocket"
            data_formats = ["tr_json2"]

        info = self.info_from_url(transport, url, data_formats=data_formats)

        return PlatformStreamCxnConfig(info, "OMM", transport=transport)


class PlatformAndDeployedCxnConfigProvider(
    DeployedCxnConfigProvider, PlatformCxnConfigProvider
):
    def get_cfg(self, session: "PlatformSession", api_cfg_key: str) -> StreamCxnConfig:
        if api_cfg_key.startswith("apis.streaming.pricing.endpoints.main"):
            cxn_config = DeployedCxnConfigProvider.get_cfg(self, session, api_cfg_key)

        else:
            cxn_config = PlatformCxnConfigProvider.get_cfg(self, session, api_cfg_key)

        return cxn_config


provider_class_by_session_cxn_type: Dict[SessionCxnType, Type[CxnConfigProvider]] = {
    SessionCxnType.DEPLOYED: DeployedCxnConfigProvider,
    SessionCxnType.REFINITIV_DATA: PlatformCxnConfigProvider,
    SessionCxnType.REFINITIV_DATA_AND_DEPLOYED: PlatformAndDeployedCxnConfigProvider,
    SessionCxnType.DESKTOP: DesktopCxnConfigProvider,
}

cache_provider_by_session = {}


def get_cxn_config(api_config_key: str, session: "Session") -> StreamCxnConfig:
    cxn_cfg_provider = get_cxn_cfg_provider(session)
    return cxn_cfg_provider.get_cfg(session, api_config_key)


def get_cxn_cfg_provider(session: "Session") -> CxnConfigProvider:
    session_cxn_type = session._get_session_cxn_type()
    provider_class = provider_class_by_session_cxn_type.get(session_cxn_type)

    if not provider_class:
        raise ValueError(
            f"Can't find provider_class by session_cxn_type={session_cxn_type}"
        )

    return cache_provider_by_session.setdefault(session, provider_class())


def release_cxn_cfg_provider(session: "Session") -> None:
    cache_provider_by_session.pop(session, None)



import abc

from ._custom_instruments_data_provider import (
    simple_custom_insts_body_params_config,
    custom_inst_type_arg_parser,
)
from ._enums import CustomInstrumentTypes
from ._instrument_prop_classes import Basket, UDC
from ..._content_type import ContentType
from ..._core.session import Session
from ...delivery._data._data_provider import DataProviderLayer
from ...delivery._data._endpoint_data import RequestMethod


class Instrument(abc.ABC):
    def __init__(self, data: dict, session: "Session" = None):
        """
        symbol : str
            Instrument symbol in the format "S)someSymbol.YOURUUID" or "someSymbol"
        formula : str, optional
            Formula consisting of rics (fields can be specified by comma).
        basket : dict, Basket, optional
            For weighted baskets / indices.
        udc : dict, UDC, optional
            Custom trading sessions, see sample format below.
        instrument_name : str, optional
            Human-readable name of the instrument. Maximum of 16 characters.
        exchange_name : str, optional
            4-letter code of the listing exchange.
        currency : str, optional
            3-letter code of the currency of the instrument, e.g. GBP.
        time_zone: str, optional
            Time Series uses an odd custom 3-letter value for time zone IDs, e.g. "LON" for London.
        holidays : List[Union[dict, Holiday]], optional
            List of custom calendar definitions.
        description : str, optional
            Free text field from the user to put any notes or text. Up to 1000 characters.
        session : Session, optional
            session=None - means default session would be used

        """
        self._data = data
        self._session = session

    @property
    def symbol(self):
        return self._data.get("symbol")

    @symbol.setter
    def symbol(self, value):
        self._data["symbol"] = value

    @property
    def instrument_name(self):
        return self._data.get("instrumentName")

    @instrument_name.setter
    def instrument_name(self, value):
        self._data["instrumentName"] = value

    @property
    def exchange_name(self):
        return self._data.get("exchangeName")

    @exchange_name.setter
    def exchange_name(self, value):
        self._data["exchangeName"] = value

    @property
    def currency(self):
        return self._data.get("currency")

    @currency.setter
    def currency(self, value):
        self._data["currency"] = value

    @property
    def time_zone(self):
        return self._data.get("timeZone")

    @time_zone.setter
    def time_zone(self, value):
        self._data["timeZone"] = value

    @property
    def holidays(self):
        return self._data.get("holidays")

    @holidays.setter
    def holidays(self, value):
        self._data["holidays"] = value

    @property
    def description(self):
        return self._data.get("description")

    @description.setter
    def description(self, value):
        self._data["description"] = value

    @property
    def id(self):
        return self._data.get("id")

    @property
    def owner(self):
        return self._data.get("owner")

    @property
    def type_(self):
        return self._data.get("type")

    def delete(self):
        """
        Examples
        --------
        >>> from refinitiv.data.content.custom_instruments.manage import get
        >>> instrument = get("MyInstrument")
        >>> instrument.delete()
        """
        data_provider_layer = DataProviderLayer(
            data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
            universe=self.symbol,
            method=RequestMethod.DELETE,
        )
        data_provider_layer.get_data(self._session)

    def save(self):
        """
        Examples:
        --------
        >>> from refinitiv.data.content.custom_instruments.manage import create_formula, Holiday
        >>> instrument = create_formula(
        ...     symbol="MyNewInstrument",
        ...     formula="EUR=*3",
        ...     holidays=[
        ...         Holiday(date="1991-08-23", name="Independence Day of Ukraine"),
        ...         {"date": "2022-12-18", "reason": "Hanukkah"},
        ...     ],
        ... )
        ... instrument.currency = "GBP"
        ... instrument.description = "short trading instrument"
        ... instrument.exchange_name = "9978"
        ... instrument.save()
        """
        provider = DataProviderLayer(
            data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
            universe=self.id,
            body_params_config=simple_custom_insts_body_params_config,
            method=RequestMethod.PUT,
            **self._data,
        )
        response = provider.get_data(self._session)
        self._data = response.data.raw


class CustomInstrumentFormula(Instrument):
    @property
    def formula(self):
        return self._data.get("formula")

    @formula.setter
    def formula(self, value):
        self._data["formula"] = value


class CustomInstrumentBasket(Instrument):
    @property
    def basket(self):
        return Basket._from_dict(self._data.get("basket"))

    @basket.setter
    def basket(self, value):
        self._data["basket"] = value


class CustomInstrumentUDC(Instrument):
    @property
    def udc(self) -> UDC:
        return UDC._from_dict(self._data.get("udc"))

    @udc.setter
    def udc(self, value):
        self._data["udc"] = value


class_by_type = {
    CustomInstrumentTypes.Formula: CustomInstrumentFormula,
    CustomInstrumentTypes.Basket: CustomInstrumentBasket,
    CustomInstrumentTypes.UDC: CustomInstrumentUDC,
}


def create_instr_factory(data, session):
    _type = data.get("type")

    if not _type:
        raise AttributeError("type parameter is not existed for object")

    enum_type = custom_inst_type_arg_parser.get_enum(_type)
    _class = class_by_type.get(enum_type)

    if not _class:
        raise AttributeError(f"There is no valid class for {enum_type}")

    ci = _class(data, session=session)
    return ci



from typing import Optional, TYPE_CHECKING

from ._constituents_description import (
    CrossCurrencyConstituentsDescription,
)
from ._curve_parameters import CrossCurrencyCurveParameters
from ...._object_definition import ObjectDefinition


if TYPE_CHECKING:
    from ......_types import OptStr


class CrossCurrencyInstrumentsSegment(ObjectDefinition):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    start_date : str, optional

    constituents : CrossCurrencyConstituentsDescription, optional

    curve_parameters : CrossCurrencyCurveParameters, optional

    """

    def __init__(
        self,
        start_date: "OptStr" = None,
        constituents: Optional[CrossCurrencyConstituentsDescription] = None,
        curve_parameters: Optional[CrossCurrencyCurveParameters] = None,
    ) -> None:
        super().__init__()
        self.start_date = start_date
        self.constituents = constituents
        self.curve_parameters = curve_parameters

    @property
    def constituents(self):
        """
        :return: object CrossCurrencyConstituentsDescription
        """
        return self._get_object_parameter(
            CrossCurrencyConstituentsDescription, "constituents"
        )

    @constituents.setter
    def constituents(self, value):
        self._set_object_parameter(
            CrossCurrencyConstituentsDescription, "constituents", value
        )

    @property
    def curve_parameters(self):
        """
        :return: object CrossCurrencyCurveParameters
        """
        return self._get_object_parameter(
            CrossCurrencyCurveParameters, "curveParameters"
        )

    @curve_parameters.setter
    def curve_parameters(self, value):
        self._set_object_parameter(
            CrossCurrencyCurveParameters, "curveParameters", value
        )

    @property
    def start_date(self):
        """
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        self._set_parameter("startDate", value)



from types import SimpleNamespace
from unittest.mock import patch

import pytest

from .conftest import raise_

package = {
    "bucket": "ESG",
    "name": "RFT-ESG-Scores-Wealth-Standard",
    "download": {
        "path": "./downloads/esg/standard_scores",
        "auto-extract": False,
        "auto-retry": {"enabled": False},
    },
}
logger = SimpleNamespace()
logger.log = lambda *_: None
logger.warning = lambda *_: None
logger.error = lambda *_: None
logger.debug = lambda *_: None
logger.info = lambda *_: None


def test_clean_up_files(file_manager):
    # given
    remove_file_mock = patch("os.remove", new=lambda *_: _)

    # when
    remove_file_mock.start()
    result = file_manager.cleanup_files()
    remove_file_mock.stop()

    # then
    assert result is None


def test_clean_up_files_if_permissions_error(file_manager):
    # given
    remove_file_mock = patch("os.remove", new=lambda *_: raise_(PermissionError))

    # when
    remove_file_mock.start()
    result = file_manager.cleanup_files()
    remove_file_mock.stop()

    # then
    assert result is None


def test_file_manager_read_file(file_manager):
    # given
    mock_path_exists = patch("os.path.exists", new=lambda: False)

    # when
    # then
    with pytest.raises(FileNotFoundError):
        file_manager.read_file("filename")



from tests.unit.conftest import StubResponse


EXPECTED_CONSTITUENTS_LIST = ["MT.AS", "ESLX.PA"]
EXPECTED_SUMMARY_LIST = [".FCHI", ".DJI", "/.STOXX50E", ".AD.FCHI", "EUR="]
BUNCH_CONSTITUENTS = [
    StubResponse(
        {
            "meta": {
                "nextLink": "bGltaXQ9MSZjdXJzb3I9MSMuRkNISQ==",
                "prevLink": "",
            },
            "universe": {"ric": "0#.FCHI"},
            "data": {"constituents": ["MT.AS"] + EXPECTED_SUMMARY_LIST},
        }
    ),
    StubResponse(
        {
            "meta": {
                "nextLink": "",
                "prevLink": "bGltaXQ9MSZjdXJzb3I9MCMuRkNISQ==",
            },
            "universe": {"ric": "1#.FCHI"},
            "data": {"constituents": ["ESLX.PA"]},
        }
    ),
]



import datetime
import uuid
from time import sleep

import pytest

import refinitiv.data as rd
from refinitiv.data import OpenState
from refinitiv.data.content import custom_instruments as ci
from refinitiv.data.content.ipa import dates_and_calendars
from tests.integration.conftest import create_desktop_session, create_platform_session_with_rdp_creds

instrument_data = {
    "description": "Glory tale about instrument",
    "exchange_name": "8080",
    "instrument_name": "Trading St. tool",
    "time_zone": "LON",
    "currency": "GBP",
}

new_instrument_data = {
    "description": "Something new with trading tool",
    "exchange_name": "9832",
    "instrument_name": "GMBH fintech t.",
    "time_zone": "LON",
    "currency": "EUR",
}

volume_based_udc = ci.manage.UDC(
    root="CC",
    months=ci.manage.Months(
        number_of_years=3,
        include_all_months=True,
        start_month=1,
    ),
    rollover=ci.manage.VolumeBasedRollover(
        method=ci.VolumeBasedRolloverMethod.VOLUME,
        number_of_days=1,
        join_at_day=1,
        roll_occurs_within_months=4,
        roll_on_expiry=True,
    ),
    spread_adjustment=ci.manage.SpreadAdjustment(
        adjustment="arithmetic",
        method=ci.SpreadAdjustmentMethod.CLOSE_TO_CLOSE,
        backwards=True,
    ),
)

day_based_udc = ci.manage.UDC(
    root="CC",
    months=ci.manage.Months(number_of_years=3, include_months=[1, 2, 3], start_month=2),
    rollover=ci.manage.DayBasedRollover(
        method=ci.DayBasedRolloverMethod.DAYS_BEFORE_END_OF_MONTH,
        number_of_days=3,
        months_prior=1,
    ),
    spread_adjustment=ci.manage.SpreadAdjustment(
        adjustment="arithmetic",
        method=ci.SpreadAdjustmentMethod.CLOSE_TO_CLOSE,
        backwards=True,
    ),
)

new_day_based_udc = ci.manage.UDC(
    root="CC",
    months=ci.manage.Months(
        number_of_years=2, include_months=[4, 5, 6, 7], start_month=3
    ),
    rollover=ci.manage.DayBasedRollover(
        method=ci.DayBasedRolloverMethod.DAYS_BEFORE_EXPIRY,
        number_of_days=2,
    ),
    spread_adjustment=ci.manage.SpreadAdjustment(
        adjustment="arithmetic",
        method=ci.SpreadAdjustmentMethod.OPEN_TO_OPEN,
        backwards=True,
    ),
)

manual_udc = ci.manage.UDC(
    root="CC",
    rollover=ci.manage.ManualRollover(
        ci.manage.ManualItem(month=7, year=2022, start_date="2022-02-01"),
        ci.manage.ManualItem(month=7, year=2021, start_date=datetime.date(2021, 3, 1)),
        ci.manage.ManualItem(
            month=3, year=2020, start_date=datetime.timedelta(days=-1550)
        ),
    ),
    spread_adjustment=ci.manage.SpreadAdjustment(
        adjustment="arithmetic",
        method="close-to-close",
        backwards=True,
    ),
)


def check_open_and_close_stream_state(
    stream, wait_before_close=None, with_updates=True
):
    stream.open(with_updates)
    assert stream.open_state == OpenState.Opened, "Stream is not in opened state"
    if wait_before_close is not None:
        sleep(wait_before_close)
    stream.close()
    assert stream.open_state == OpenState.Closed, "Stream is not in closed state"


def check_close_and_open_stream_state(stream_01, stream_02, stream_03, wait_time=None):
    stream_01.close()
    assert stream_01.open_state == OpenState.Closed, "Stream is not in closed state"
    if wait_time is not None:
        sleep(wait_time)
    assert stream_02.open_state == OpenState.Opened, "Stream is not in opened state"
    assert stream_03.open_state == OpenState.Opened, "Stream is not in opened state"


def check_stream_instrument(expected_instrument, stream):
    if isinstance(expected_instrument, list):
        actual_universe = stream._stream.universe
    else:
        actual_universe = stream._stream.universe[0]
    assert (
        expected_instrument == actual_universe
    ), f"{actual_universe} universe is not expected"


def is_stream_id_equal(stream_01, stream_02):
    return stream_02._stream._id == stream_01._stream._id


def check_stream_snapshot(stream, universe, expected_fields):
    df = stream.get_snapshot()
    assert not df.empty, f"Snapshot dataframe for {universe} is empty"
    assert universe in df.values[0], f"There is no data for {universe} in snapshot"
    actual_fields = set(df.columns)
    assert (
        actual_fields == set(expected_fields),
    ), f"{expected_fields} != {actual_fields}"


def get_holidays(session):
    calendar_holiday = dates_and_calendars.holidays.Definition(
        start_date="2015-08-24",
        end_date="2018-09-24",
        calendars=["UKR"],
        holiday_outputs=["Date", "Names"],
    ).get_data(session=session)
    return calendar_holiday.data.holidays[0]


@pytest.fixture(name="symbol")
def get_ci_symbol():
    return str(uuid.uuid4())[:8]


@pytest.fixture(
    scope="function",
    params=[create_platform_session_with_rdp_creds, create_desktop_session],
    ids=["platform_session", "desktop_session"],
)
def open_session_with_rdp_creds_for_ci(request, set_env_base_url):
    create_session = request.param
    session = create_session()
    rd.session.set_default(session)
    session.open()
    yield session
    session.close()
    rd.session.set_default(None)


import string
from functools import lru_cache
from typing import Set, Tuple

from jinja2 import Environment, nodes

from refinitiv.data._tools._common import cached_property


class OldStyleStringTemplate(string.Template):
    """string.Template with extended validation capabilities

    This template must be used directly or as a base for templates in refinitiv.data
    JSON configuration file.
    Default delimiter is '#'. Variable syntax: '#var' or '#{var}'.
    """

    delimiter = "#"
    braceidpattern = r"(?a:[_a-z][_.|a-z0-9]*)"

    def names(self) -> Set[str]:
        """Get names of substitution variables in pattern"""
        return {
            match_tuple[1] or match_tuple[2]  # named or braced
            for match_tuple in self.pattern.findall(self.template)
            if match_tuple[1] or match_tuple[2]
        }


class StringTemplate:
    """Jinja-based string template"""

    def __init__(self, source):
        self.env = Environment(
            block_start_string="{{",
            block_end_string="}}",
            variable_start_string="#{",
            variable_end_string="}",
        )

        self.env.filters["in"] = lambda inp: ", ".join(map(repr, inp))
        self.source = source

    @property
    def template(self):
        return self.source

    @staticmethod
    @lru_cache(128)
    def get_vars_from_node(node: nodes.Node) -> Set[str]:
        vars_set = set()
        iter_list = [(node,)]
        while iter_list:
            _node, *_attrs = iter_list.pop(0)
            if isinstance(_node, nodes.Getattr):
                _attrs = (_node.attr, *_attrs)
            elif isinstance(_node, nodes.Name):
                vars_set.add(".".join(map(str, (_node.name, *_attrs))))
            iter_list.extend((i, *_attrs) for i in _node.iter_child_nodes())
        return vars_set

    def placeholders(self):
        return self.get_vars_from_node(self._parsed)

    @cached_property
    def _parsed(self):
        return self.env.parse(self.source)

    def validate(self):
        self._parsed

    def substitute(self, **kwargs):
        return self.env.from_string(self.source).render(**kwargs)


class InvalidPlaceholderError(ValueError):
    """Exception to display syntax errors in string templates"""

    def __init__(
        self,
        index: int,
        template_text: str,
        limit: int = 80,
        padding: int = 0,
        prefix: str = "",
    ):
        """
        Parameters
        ----------
        index : int
            index of wrong placeholder start, get it from regex match
        template_text
            original text template
        limit : int
            maximum length of error message
        padding : int
            padding around place where invalid placeholder starts in error message text
        prefix : int
            prefix before error message, for additional information
        """
        self.limit = limit
        self.padding = padding
        self.index = index
        self.template_text = template_text
        self.prefix = prefix

    def __str__(self):
        line_index, col_index = index_to_line_and_col(self.index, self.template_text)
        target_line = self.template_text.splitlines()[line_index]
        target_line, shift = shorten_string_to_include_position(
            target_line, self.limit, col_index, self.padding
        )
        return "\n".join(
            [
                f"{self.prefix}Invalid placeholder in the template string: "
                f"line {line_index + 1}, col {col_index + 1}:",
                target_line,
                "-" * (col_index - shift) + "^",
            ]
        )


def index_to_line_and_col(index: int, target_string: str) -> Tuple[int, int]:
    """Convert position index in multiline string to line and column

    Parameters
    ----------
    index : int
        index of symbol in target string
    target_string : str
        target string

    Returns
    -------
    Tuple of line index and column index of given symbol, starting from zero
    """
    lines = target_string[:index].splitlines(keepends=True)

    if not lines:
        return 0, 0

    col_index = index - len("".join(lines[:-1]))
    line_index = len(lines) - 1
    return line_index, col_index


def shorten_string_to_include_position(
    line: str, limit: int, pos: int, padding: int = 0
) -> Tuple[str, int]:
    """Shorten string to given limit to include given position

    Can be used when we need to display position in a string when screen width is
    limited.

    Parameters
    ----------
    line : str
        target string
    limit : int
        maximum length of resulting string
    pos : int
        position in string that must be included in shortened string, starting from 0
    padding
        number of symbols left and right of pos that also must be included

    Returns
    -------
    Tuple of shortened string and number of symbols removed from the start
    """
    cur_right = pos + padding
    if padding >= len(line):
        raise ValueError("padding must be less than the length of line")
    if len(line) <= limit:
        return line, 0
    if cur_right < limit:
        return line[:limit], 0
    left = max(cur_right - limit + 1, 0)
    return line[left : cur_right + 1], left



import pytest

import refinitiv.data.content as rdc
from refinitiv.data.content.ipa.curves import zc_curve_definitions
from refinitiv.data.content.ipa.surfaces import cap, eti, fx
from refinitiv.data.delivery._data._data_provider_factory import make_provider
from tests.unit.conftest import StubSession


@pytest.mark.parametrize(
    "definition, expected_value",
    [
        (
            rdc.ipa.curves.forward_curves.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {"universe": [{"my_test_param": "test_value"}]},
        ),
        (
            rdc.ipa.curves.forward_curves.Definitions(
                [
                    rdc.ipa.curves.forward_curves.Definition(
                        extended_params={"my_test_param1": "test_value1"}
                    )
                ],
            ),
            {
                "universe": [{"my_test_param1": "test_value1"}],
            },
        ),
        (
            rdc.ipa.curves.zc_curves.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {"universe": [{"my_test_param": "test_value"}]},
        ),
        (
            rdc.ipa.curves.zc_curve_definitions.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {"universe": [{"my_test_param": "test_value"}]},
        ),
        (
            rdc.ipa.curves.zc_curve_definitions.Definitions(
                [
                    zc_curve_definitions.Definition(
                        source="Refinitiv",
                        extended_params={"my_test_param1": "test_value1"},
                    ),
                ],
            ),
            {
                "universe": [{"source": "Refinitiv", "my_test_param1": "test_value1"}],
            },
        ),
        (
            rdc.ipa.curves.zc_curves.Definitions(
                [
                    rdc.ipa.curves.zc_curve_definitions.Definition(
                        extended_params={"my_test_param1": "test_value1"}
                    )
                ],
            ),
            {
                "universe": [{"my_test_param1": "test_value1"}],
            },
        ),
        (
            rdc.ipa.financial_contracts.bond.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {
                "universe": [
                    {
                        "instrumentDefinition": {},
                        "instrumentType": "Bond",
                        "my_test_param": "test_value",
                    }
                ]
            },
        ),
        (
            rdc.ipa.financial_contracts.cap_floor.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {
                "universe": [
                    {
                        "instrumentDefinition": {},
                        "instrumentType": "CapFloor",
                        "my_test_param": "test_value",
                    }
                ]
            },
        ),
        (
            rdc.ipa.financial_contracts.cds.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {
                "universe": [
                    {
                        "instrumentDefinition": {},
                        "instrumentType": "Cds",
                        "my_test_param": "test_value",
                    }
                ]
            },
        ),
        (
            rdc.ipa.financial_contracts.cross.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {
                "universe": [
                    {
                        "instrumentDefinition": {},
                        "instrumentType": "FxCross",
                        "my_test_param": "test_value",
                    }
                ]
            },
        ),
        (
            rdc.ipa.financial_contracts.option.Definition(extended_params={}),
            {
                "universe": [
                    {
                        "instrumentDefinition": {},
                        "instrumentType": "Option",
                    }
                ]
            },
        ),
        (
            rdc.ipa.financial_contracts.repo.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {
                "universe": [
                    {
                        "instrumentDefinition": {},
                        "instrumentType": "Repo",
                        "my_test_param": "test_value",
                    }
                ]
            },
        ),
        (
            rdc.ipa.financial_contracts.swap.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {
                "universe": [
                    {
                        "instrumentDefinition": {},
                        "instrumentType": "Swap",
                        "my_test_param": "test_value",
                    }
                ]
            },
        ),
        (
            rdc.ipa.financial_contracts.swaption.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {
                "universe": [
                    {
                        "instrumentDefinition": {},
                        "instrumentType": "Swaption",
                        "my_test_param": "test_value",
                    }
                ]
            },
        ),
        (
            rdc.ipa.financial_contracts.term_deposit.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {
                "universe": [
                    {
                        "instrumentDefinition": {},
                        "instrumentType": "TermDeposit",
                        "my_test_param": "test_value",
                    }
                ]
            },
        ),
        (
            rdc.ipa.surfaces.cap.Definition(
                surface_layout=cap.SurfaceLayout(format=cap.Format.MATRIX),
                surface_parameters=cap.CapCalculationParams(
                    valuation_date="2020-03-20",
                    x_axis=cap.Axis.STRIKE,
                    y_axis=cap.Axis.TENOR,
                ),
                underlying_definition=cap.CapSurfaceDefinition(
                    instrument_code="USD123",
                    discounting_type=cap.DiscountingType.OIS_DISCOUNTING,
                ),
                surface_tag="USD_Strike__Tenor_",
                extended_params={"my_test_param": "test_value"},
            ),
            {
                "universe": [
                    {
                        "my_test_param": "test_value",
                        "surfaceLayout": {"format": "Matrix"},
                        "surfaceParameters": {
                            "valuationDate": "2020-03-20",
                            "xAxis": "Strike",
                            "yAxis": "Tenor",
                        },
                        "surfaceTag": "USD_Strike__Tenor_",
                        "underlyingDefinition": {
                            "discountingType": "OisDiscounting",
                            "instrumentCode": "USD123",
                        },
                        "underlyingType": "Cap",
                    }
                ]
            },
        ),
        (
            rdc.ipa.surfaces.eti.Definition(
                surface_layout=eti.SurfaceLayout(
                    format=eti.Format.MATRIX, y_point_count=10
                ),
                surface_parameters=eti.EtiCalculationParams(
                    price_side=eti.PriceSide.MID,
                    volatility_model=eti.VolatilityModel.SVI,
                    x_axis=eti.Axis.DATE,
                    y_axis=eti.Axis.STRIKE,
                ),
                underlying_definition={"instrumentCode": "BNPP.PA@RIC"},
                surface_tag="1",
                extended_params={"my_test_param": "test_value"},
            ),
            {
                "universe": [
                    {
                        "my_test_param": "test_value",
                        "surfaceLayout": {"format": "Matrix", "yPointCount": 10},
                        "surfaceParameters": {
                            "priceSide": "Mid",
                            "volatilityModel": "SVI",
                            "xAxis": "Date",
                            "yAxis": "Strike",
                        },
                        "surfaceTag": "1",
                        "underlyingDefinition": {"instrumentCode": "BNPP.PA@RIC"},
                        "underlyingType": "Eti",
                    }
                ]
            },
        ),
        (
            rdc.ipa.surfaces.fx.Definition(
                surface_layout=fx.SurfaceLayout(format=fx.Format.MATRIX),
                surface_parameters=fx.FxCalculationParams(
                    x_axis=fx.Axis.DATE,
                    y_axis=fx.Axis.STRIKE,
                    calculation_date="2018-08-20T00:00:00Z",
                ),
                underlying_definition={"fxCrossCode": "EURUSD"},
                extended_params={"my_test_param": "test_value"},
            ),
            {
                "universe": [
                    {
                        "my_test_param": "test_value",
                        "surfaceLayout": {"format": "Matrix"},
                        "surfaceParameters": {
                            "calculationDate": "2018-08-20T00:00:00Z",
                            "xAxis": "Date",
                            "yAxis": "Strike",
                        },
                        "underlyingDefinition": {"fxCrossCode": "EURUSD"},
                        "underlyingType": "Fx",
                    }
                ]
            },
        ),
        (
            rdc.ipa.surfaces.swaption.Definition(
                extended_params={"my_test_param": "test_value"}
            ),
            {
                "universe": [
                    {"my_test_param": "test_value", "underlyingType": "Swaption"}
                ]
            },
        ),
    ],
    ids=[
        "rdc.ipa.curves.forward_curves.Definition",
        "rdc.ipa.curves.forward_curves.Definitions",
        "rdc.ipa.curves.zc_curves.Definition",
        "rdc.ipa.curves.zc_curve_definitions.Definition",
        "rdc.ipa.curves.zc_curve_definitions.Definitions",
        "rdc.ipa.curves.zc_curves.Definitions",
        "rdc.ipa.financial_contracts.bond.Definition",
        "rdc.ipa.financial_contracts.cap_floor.Definition",
        "rdc.ipa.financial_contracts.cds.Definition",
        "rdc.ipa.financial_contracts.cross.Definition",
        "rdc.ipa.financial_contracts.option.Definition",
        "rdc.ipa.financial_contracts.repo.Definition",
        "rdc.ipa.financial_contracts.swap.Definition",
        "rdc.ipa.financial_contracts.swaption.Definition",
        "rdc.ipa.financial_contracts.term_deposit.Definition",
        "rdc.ipa.surfaces.cap.Definition",
        "rdc.ipa.surfaces.eti.Definition",
        "rdc.ipa.surfaces.fx.Definition",
        "rdc.ipa.surfaces.swaption.Definition",
    ],
)
def test_extended_params(definition, expected_value):
    content_type = definition._kwargs["__content_type__"]
    provider = make_provider(content_type)

    # when
    request = provider.request.create(StubSession(), "url", **definition._kwargs)

    # then
    testing_value = request.json
    assert testing_value == expected_value, definition



# coding: utf8

from typing import Optional

from ._enums import RepoCurveType
from ..._object_definition import ObjectDefinition


class PricingParameters(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    repo_curve_type : RepoCurveType, optional
        Curve used to compute the repo rate. it can be computed using following methods:
        - repocurve : rate is computed by interpolating a repo curve.     - depositcurve
        : rate is computed by interpolating a deposit curve.     - fixinglibor : rate is
        computed by interpolating libor rates.  if no curve can be found, the rate is
        computed using a deposit curve.
    market_data_date : str, optional
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
    report_ccy : str, optional
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
    valuation_date : str, optional
        The date at which the instrument is valued. the value is expressed in iso 8601
        format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). by default,
        marketdatadate is used. if marketdatadate is not specified, the default value is
        today.
    """

    def __init__(
        self,
        repo_curve_type: Optional[RepoCurveType] = None,
        market_data_date: Optional[str] = None,
        report_ccy: Optional[str] = None,
        valuation_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.repo_curve_type = repo_curve_type
        self.market_data_date = market_data_date
        self.report_ccy = report_ccy
        self.valuation_date = valuation_date

    @property
    def repo_curve_type(self):
        """
        Curve used to compute the repo rate. it can be computed using following methods:
        - repocurve : rate is computed by interpolating a repo curve.     - depositcurve
        : rate is computed by interpolating a deposit curve.     - fixinglibor : rate is
        computed by interpolating libor rates.  if no curve can be found, the rate is
        computed using a deposit curve.
        :return: enum RepoCurveType
        """
        return self._get_enum_parameter(RepoCurveType, "repoCurveType")

    @repo_curve_type.setter
    def repo_curve_type(self, value):
        self._set_enum_parameter(RepoCurveType, "repoCurveType", value)

    @property
    def market_data_date(self):
        """
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def report_ccy(self):
        """
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
        :return: str
        """
        return self._get_parameter("reportCcy")

    @report_ccy.setter
    def report_ccy(self, value):
        self._set_parameter("reportCcy", value)

    @property
    def valuation_date(self):
        """
        The date at which the instrument is valued. the value is expressed in iso 8601
        format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). by default,
        marketdatadate is used. if marketdatadate is not specified, the default value is
        today.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



# coding: utf8

from enum import Enum, unique


@unique
class RoundingType(Enum):
    CEIL = "Ceil"
    DEFAULT = "Default"
    DOWN = "Down"
    FACE_DOWN = "FaceDown"
    FACE_NEAR = "FaceNear"
    FACE_UP = "FaceUp"
    FLOOR = "Floor"
    NEAR = "Near"
    UP = "Up"



# coding: utf8

from enum import Enum, unique


@unique
class BusinessDayConvention(Enum):
    BBSW_MODIFIED_FOLLOWING = "BbswModifiedFollowing"
    EVERY_THIRD_WEDNESDAY = "EveryThirdWednesday"
    MODIFIED_FOLLOWING = "ModifiedFollowing"
    NEXT_BUSINESS_DAY = "NextBusinessDay"
    NO_MOVING = "NoMoving"
    PREVIOUS_BUSINESS_DAY = "PreviousBusinessDay"



from typing import TYPE_CHECKING

from ._request import Request

if TYPE_CHECKING:
    import httpx
    from ..._core.session import Session


class HttpSessionConnection:
    def send(
        self, request: Request, session: "Session", *args, auto_retry=False, **kwargs
    ) -> "httpx.Response":

        request.auto_retry = auto_retry

        return session.http_request(request)

    async def send_async(
        self, request: Request, session: "Session", *args, auto_retry=False, **kwargs
    ) -> "httpx.Response":

        request.auto_retry = auto_retry

        return await session.http_request_async(request)



from enum import Enum, unique


@unique
class EconomicSector(Enum):
    ACADEMIC_AND_EDUCATIONAL_SERVICES = "AcademicAndEducationalServices"
    BASIC_MATERIALS = "BasicMaterials"
    CONSUMER_CYCLICALS = "ConsumerCyclicals"
    CONSUMER_NON_CYCLICALS = "ConsumerNonCyclicals"
    ENERGY = "Energy"
    FINANCIALS = "Financials"
    GOVERNMENT_ACTIVITY = "GovernmentActivity"
    HEALTHCARE = "Healthcare"
    INDUSTRIALS = "Industrials"
    INSTITUTIONS_ASSOCIATIONS_AND_ORGANIZATIONS = (
        "InstitutionsAssociationsAndOrganizations"
    )
    REAL_ESTATE = "RealEstate"
    TECHNOLOGY = "Technology"
    UTILITIES = "Utilities"



from typing import TYPE_CHECKING, Iterable, Callable, List

from ._data_grid_type import use_field_names_in_headers_arg_parser
from .._content_data_provider import ContentDataProvider
from .._content_response_factory import ContentResponseFactory
from ..._tools import (
    universe_arg_parser,
    fields_arg_parser,
    ADC_TR_PATTERN,
    ADC_FUNC_PATTERN_IN_FIELDS,
    cached_property,
)
from ...delivery._data import RequestMethod
from ...delivery._data._data_provider import (
    RequestFactory,
    ContentValidator,
    ValidatorContainer,
    Request,
)

if TYPE_CHECKING:
    from ...delivery._data._data_provider import ParsedData

# --------------------------------------------------------------------------------------
#   Response
# --------------------------------------------------------------------------------------
error_message_by_code = {
    "default": "{error_message} Requested universes: {universes}. Requested fields: {"
    "fields}",
    412: "Unable to resolve all requested identifiers in {universes}.",
    218: "Unable to resolve all requested fields in {fields}. The formula must "
    "contain at least one field or function.",
}


class DataGridResponseFactory(ContentResponseFactory):
    def create_fail(
        self, parsed_data: "ParsedData", universe=None, fields=None, **kwargs
    ):
        error_code = parsed_data.first_error_code
        if error_code not in error_message_by_code.keys():
            parsed_data.error_messages = error_message_by_code["default"].format(
                error_message=parsed_data.first_error_message,
                fields=fields,
                universes=universe,
            )
        else:
            parsed_data.error_messages = error_message_by_code[error_code].format(
                fields=fields, universes=universe
            )

        return super().create_fail(parsed_data, **kwargs)


class DataGridRDPResponseFactory(DataGridResponseFactory):
    def create_success(self, parsed_data: "ParsedData", **kwargs):
        inst = super().create_success(parsed_data, **kwargs)
        descriptions = (
            self.get_raw(parsed_data).get("messages", {}).get("descriptions", [])
        )
        for descr in descriptions:
            code = descr.get("code")
            if code in {416, 413}:
                inst.errors.append((code, descr.get("description")))

        return inst


class DataGridUDFResponseFactory(DataGridResponseFactory):
    def get_raw(self, parsed_data: "ParsedData"):
        return parsed_data.content_data.get("responses", [{}])[0]

    def create_success(self, parsed_data: "ParsedData", **kwargs):
        inst = super().create_success(parsed_data, **kwargs)
        error = self.get_raw(parsed_data).get("error", [])
        for err in error:
            code = err.get("code")
            if code == 416:
                inst.errors.append((code, err.get("message")))

        return inst


# --------------------------------------------------------------------------------------
#   Request
# --------------------------------------------------------------------------------------


def validate_correct_format_parameters(*_, **kwargs) -> dict:
    parameters = kwargs.get("parameters")
    extended_params = kwargs.get("extended_params")
    universe = kwargs.get("universe")
    fields = kwargs.get("fields")
    use_field_names_in_headers = kwargs.get("use_field_names_in_headers")

    if parameters is not None and not isinstance(parameters, dict):
        raise ValueError(f"Arg parameters must be a dictionary")

    extended_params = extended_params or {}
    universe = extended_params.get("universe") or universe
    universe = universe_arg_parser.get_list(universe)
    universe = [value.upper() if value.islower() else value for value in universe]
    fields = fields_arg_parser.get_list(fields)
    use_field_names_in_headers = use_field_names_in_headers_arg_parser.get_bool(
        use_field_names_in_headers
    )

    kwargs.update(
        {
            "universe": universe,
            "fields": fields,
            "parameters": parameters,
            "use_field_names_in_headers": use_field_names_in_headers,
            "extended_params": extended_params,
        }
    )
    return kwargs


class DataGridRDPRequestFactory(RequestFactory):
    def get_body_parameters(self, *_, **kwargs) -> dict:
        kwargs = validate_correct_format_parameters(*_, **kwargs)
        body_parameters = {}

        universe = kwargs.get("universe")
        if universe:
            body_parameters["universe"] = universe

        fields = kwargs.get("fields")
        if fields:
            body_parameters["fields"] = fields

        parameters = kwargs.get("parameters")
        if parameters:
            body_parameters["parameters"] = parameters

        layout = kwargs.get("layout")
        if isinstance(layout, dict) and layout.get("output"):
            body_parameters["output"] = layout["output"]

        return body_parameters

    def get_request_method(self, **kwargs) -> RequestMethod:
        return RequestMethod.POST


class DataGridUDFRequestFactory(RequestFactory):
    def create(self, session, *args, **kwargs):
        url_root = session._get_rdp_url_root()
        url = url_root.replace("rdp", "udf")

        method = self.get_request_method(**kwargs)
        header_parameters = kwargs.get("header_parameters") or {}
        extended_params = kwargs.get("extended_params") or {}
        body_parameters = self.get_body_parameters(*args, **kwargs)
        body_parameters = self.extend_body_parameters(body_parameters, extended_params)

        headers = {"Content-Type": "application/json"}
        headers.update(header_parameters)

        return Request(
            url=url,
            method=method,
            headers=headers,
            json={
                "Entity": {
                    "E": "DataGrid_StandardAsync",
                    "W": {"requests": [body_parameters]},
                }
            },
        )

    def get_body_parameters(self, *_, **kwargs) -> dict:
        ticket = kwargs.get("ticket", None)
        if ticket:
            return {"ticket": ticket}

        kwargs = validate_correct_format_parameters(*_, **kwargs)
        body_parameters = {}

        instruments = kwargs.get("universe")
        if instruments:
            body_parameters["instruments"] = instruments

        fields = kwargs.get("fields")
        if fields:
            body_parameters["fields"] = [
                {"name": i}
                for i in fields
                if ADC_TR_PATTERN.match(i) or ADC_FUNC_PATTERN_IN_FIELDS.match(i)
            ]

        parameters = kwargs.get("parameters")
        if parameters:
            body_parameters["parameters"] = parameters

        layout = kwargs.get("layout")
        if isinstance(layout, dict) and layout.get("layout"):
            body_parameters["layout"] = layout["layout"]

        return body_parameters

    def get_request_method(self, **kwargs) -> RequestMethod:
        return RequestMethod.POST


# --------------------------------------------------------------------------------------
#   Content data validator
# --------------------------------------------------------------------------------------


class DataGridContentValidator(ContentValidator):
    @classmethod
    def status_is_not_error(cls, data: "ParsedData") -> bool:
        status_content = data.status.get("content", "")
        if status_content.startswith("Failed"):
            data.error_codes = -1
            data.error_messages = status_content
            return False

        return True


class DataGridRDPContentValidator(DataGridContentValidator):
    @classmethod
    def content_data_has_no_error(cls, data: "ParsedData") -> bool:
        content_data = data.content_data
        error = content_data.get("error")
        if error and not content_data.get("data"):
            data.error_codes = error.get("code", -1)
            data.error_messages = error.get("description")

            if not data.error_messages:
                error_message = error.get("message")
                errors = error.get("errors")

                if isinstance(errors, list):
                    error_message += ":\n"
                    error_message += "\n".join(map(str, errors))

                data.error_messages = error_message

            return False

        return True

    @cached_property
    def validators(self) -> List[Callable[["ParsedData"], bool]]:
        return [
            self.status_is_not_error,
            self.content_data_is_not_none,
            self.content_data_has_no_error,
        ]


class DataGridUDFContentValidator(DataGridContentValidator):
    @classmethod
    def content_data_is_valid_type(cls, data: "ParsedData") -> bool:
        content_data = data.content_data
        if isinstance(content_data, str):
            data.error_codes = -1
            data.error_messages = content_data
            return False

        return True

    @classmethod
    def content_data_has_valid_response(cls, data: "ParsedData") -> bool:
        responses = data.content_data.get("responses", [])
        first_response = responses[0] if responses else {}
        error = first_response.get("error")
        if error and not first_response.get("data"):
            if isinstance(error, dict):
                data.error_codes = error.get("code", -1)
                data.error_messages = error.get("message", error)

            else:
                data.error_codes = -1
                data.error_messages = error

            return False

        return True

    @classmethod
    def content_data_response_has_valid_data(cls, data: "ParsedData") -> bool:
        responses = data.content_data.get("responses", [])
        first_response = responses[0] if responses else {}
        error = first_response.get("error")
        if error and not any(
            any(items[1:]) if isinstance(items, Iterable) else False
            for items in first_response.get("data")
        ):
            first_error = error[0]
            data.error_codes = first_error.get("code", -1)
            data.error_messages = first_error.get("message", first_error)
            return False

        return True

    @cached_property
    def validators(self) -> List[Callable[["ParsedData"], bool]]:
        return [
            self.status_is_not_error,
            self.content_data_is_not_none,
            self.content_data_is_valid_type,
            self.content_data_has_no_error,
            self.content_data_has_valid_response,
            self.content_data_response_has_valid_data,
        ]


# --------------------------------------------------------------------------------------
#   Providers
# --------------------------------------------------------------------------------------


data_grid_rdp_data_provider = ContentDataProvider(
    request=DataGridRDPRequestFactory(),
    response=DataGridRDPResponseFactory(),
    validator=ValidatorContainer(content_validator=DataGridRDPContentValidator()),
)

data_grid_udf_data_provider = ContentDataProvider(
    request=DataGridUDFRequestFactory(),
    response=DataGridUDFResponseFactory(),
    validator=ValidatorContainer(content_validator=DataGridUDFContentValidator()),
)



import allure
import pytest

import refinitiv.data as rd
import tests.integration.conftest as global_conftest
from tests.integration.access.conftest import check_session_state


@allure.suite("FinCoder layer")
@allure.feature("FinCoder - Session")
@allure.severity(allure.severity_level.CRITICAL)
class TestSession:
    @allure.title("Open, close default session and check session status")
    @pytest.mark.caseid("36971083")
    def test_open_and_close_default_session(self, load_config):
        session = rd.open_session()
        check_session_state(session)

    @allure.title(
        "Open, close default session with specific config file and check session status"
    )
    @pytest.mark.caseid("36971084")
    def test_open_and_close_default_session_with_specific_config_file(self):
        session = rd.open_session(config_name=global_conftest.conf)
        check_session_state(session)

    @allure.title(
        "Open, close specific session with specific config file and check session status"
    )
    @pytest.mark.caseid("36971085")
    def test_open_and_close_specific_session_with_specific_config_file(self):
        session = rd.open_session(
            name="platform.my-session", config_name=global_conftest.conf
        )
        check_session_state(session)

    @allure.title("Open, close session when put only app key and check session status")
    @pytest.mark.caseid("36971346")
    def test_open_and_close_when_put_app_key(self):
        session = rd.open_session(app_key=global_conftest.desktop_app_key)
        check_session_state(session)

    @allure.title("Open, close session with all params and check session status")
    @pytest.mark.caseid("36971350")
    def test_open_and_close_with_all_params(self):
        session = rd.open_session(
            name="desktop.my-session",
            app_key=global_conftest.desktop_app_key,
            config_name=global_conftest.conf,
        )
        check_session_state(session)

    @allure.title(
        "Open, close session with specific session name and check session status"
    )
    @pytest.mark.caseid("36971570")
    def test_open_and_close_with_specific_session_name(self, load_config):
        session = rd.open_session("desktop.my-session")
        check_session_state(session)

    @allure.title("Open session with invalid session name")
    @pytest.mark.caseid("36971742")
    def test_open_with_invalid_session_name(self, load_config):
        with pytest.raises(
            NameError, match="Cannot open session desktop.custom-session"
        ):
            rd.open_session("desktop.custom-session")

    @allure.title("Open session with invalid config path")
    @pytest.mark.caseid("36971748")
    def test_open_with_invalid_config_path(self):
        with pytest.raises(FileNotFoundError, match="Can't find file:"):
            rd.open_session(config_name="./path/to/fin-coder-prod.json")



import refinitiv.data.content.ipa.financial_contracts as rdf
from tests.integration.conftest import is_open

from tests.integration.constants_list import HttpStatusCode


def option_definition_02():
    definition = rdf.option.Definition(
        instrument_code="FCHI560000L5.p",
        underlying_type=rdf.option.UnderlyingType.ETI,
        underlying_definition=rdf.option.EtiUnderlyingDefinition("FCHI560000L5.p"),
        fields=[
            "MarketValueInDealCcy",
            "DeltaPercent",
            "GammaPercent",
            "RhoPercent",
            "ThetaPercent",
            "VegaPercent",
            "ErrorCode",
            "ErrorMessage",
        ],
    )

    return definition


def check_http_status_is_success_and_df_value_not_empty(response):
    status = response.http_status
    df = response.data.df
    assert status["http_status_code"] == HttpStatusCode.TWO_HUNDRED, (
        f"Actual status code is {status['http_status_code']}, "
        f"Error: {response.http_status['error']}"
    )

    assert df is not None, f"DataFrame is {df}"
    assert not df.empty, f"DataFrame is empty: {df}"
    assert not bool(df.ErrorMessage[0]), f"{df.ErrorMessage[0]}"
    assert df.GammaPercent is not None, f"{df.GammaPercent}"


def check_stream_state_and_df_from_stream(stream):
    df = stream.get_snapshot()

    assert is_open(stream), f"Stream is not open"
    assert df is not None, f"DataFrame is {df}"
    assert not df.empty, f"DataFrame is empty: {df}"
    assert df.GammaPercent is not None, f"{df.GammaPercent}"
    assert not bool(df.ErrorMessage[0]), f"{df.ErrorMessage[0]}"


option_universe = {
    "instrumentType": "Option",
    "instrumentDefinition": {
        "InstrumentTag": "my tag",
        "tenor": "5M",
        "notionalCcy": "AUD",
        "underlyingDefinition": {"fxCrossCode": "AUDUSD"},
        "underlyingType": "Fx",
        "strike": 265,
    },
    "pricingParameters": {
        "fxSpotObject": {"bid": 0.7387, "ask": 0.7387, "mid": 0.7387},
        "priceSide": "Mid",
        "pricingModelType": "BlackScholes",
        "valuationDate": "2018-08-06",
    },
}
